import sys
import os
import logging
import time
import subprocess
import zipfile
import re
import threading
import platform
import shutil

import requests
from tabulate import tabulate
from fastapi import HTTPException, status
from fastapi.responses import StreamingResponse

from openai import OpenAI

from lemonade_server.pydantic_models import (
    ChatCompletionRequest,
    PullConfig,
    EmbeddingsRequest,
    RerankingRequest,
)
from lemonade_server.model_manager import ModelManager
from lemonade.tools.server.utils.port import find_free_port

LLAMA_VERSION = "b5787"


def llamacpp_address(port: int) -> str:
    """
    Generate the base URL for the llamacpp server.

    Args:
        port: The port number the llamacpp server is running on

    Returns:
        The base URL for the llamacpp server
    """
    return f"http://127.0.0.1:{port}/v1"


def get_llama_server_paths():
    """
    Get platform-specific paths for llama server directory and executable
    """
    base_dir = os.path.join(os.path.dirname(sys.executable), "llama_server")

    if platform.system().lower() == "windows":
        return base_dir, os.path.join(base_dir, "llama-server.exe")
    else:  # Linux/Ubuntu
        # Check if executable exists in build/bin subdirectory (Current Ubuntu structure)
        build_bin_path = os.path.join(base_dir, "build", "bin", "llama-server")
        if os.path.exists(build_bin_path):
            return base_dir, build_bin_path
        else:
            # Fallback to root directory
            return base_dir, os.path.join(base_dir, "llama-server")


def get_binary_url_and_filename(version):
    """
    Get the appropriate binary URL and filename based on platform
    """
    system = platform.system().lower()

    if system == "windows":
        filename = f"llama-{version}-bin-win-vulkan-x64.zip"
    elif system == "linux":
        filename = f"llama-{version}-bin-ubuntu-vulkan-x64.zip"
    else:
        raise NotImplementedError(
            f"Platform {system} not supported for llamacpp. Supported: Windows, Ubuntu Linux"
        )

    url = (
        f"https://github.com/ggml-org/llama.cpp/releases/download/{version}/{filename}"
    )
    return url, filename


def validate_platform_support():
    """
    Validate platform support before attempting download
    """
    system = platform.system().lower()

    if system not in ["windows", "linux"]:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=(
                f"Platform {system} not supported for llamacpp. "
                "Supported: Windows, Ubuntu Linux"
            ),
        )

    if system == "linux":
        # Check if we're actually on Ubuntu/compatible distro and log a warning if not
        try:
            with open("/etc/os-release", "r", encoding="utf-8") as f:
                os_info = f.read().lower()
                if "ubuntu" not in os_info and "debian" not in os_info:
                    logging.warning(
                        "llamacpp binaries are built for Ubuntu. "
                        "Compatibility with other Linux distributions is not guaranteed."
                    )
        except (FileNotFoundError, PermissionError, OSError) as e:
            logging.warning(
                "Could not determine Linux distribution (%s). "
                "llamacpp binaries are built for Ubuntu.",
                str(e),
            )


class LlamaTelemetry:
    """
    Manages telemetry data collection and display for llama server.
    """

    def __init__(self):
        self.input_tokens = None
        self.output_tokens = None
        self.time_to_first_token = None
        self.tokens_per_second = None
        self.prompt_eval_time = None
        self.eval_time = None
        self.port = None

    def choose_port(self):
        """
        Users probably don't care what port we start llama-server on, so let's
        search for an empty port
        """

        self.port = find_free_port()

        if self.port is None:
            msg = "Failed to find an empty port to start llama-server on"
            logging.error(msg)
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail=msg,
            )

    def parse_telemetry_line(self, line: str):
        """
        Parse telemetry data from llama server output lines.
        """

        # Parse Vulkan device detection
        vulkan_match = re.search(r"ggml_vulkan: Found (\d+) Vulkan devices?:", line)
        if vulkan_match:
            device_count = int(vulkan_match.group(1))
            if device_count > 0:
                logging.info(
                    f"GPU acceleration active: {device_count} Vulkan device(s) "
                    "detected by llama-server"
                )
            return

        # Parse prompt evaluation line
        prompt_match = re.search(
            r"prompt eval time\s*=\s*([\d.]+)\s*ms\s*/\s*(\d+)\s*tokens.*?"
            r"([\d.]+)\s*tokens per second",
            line,
        )
        if prompt_match:
            prompt_time_ms = float(prompt_match.group(1))
            input_tokens = int(prompt_match.group(2))

            self.prompt_eval_time = prompt_time_ms / 1000.0
            self.input_tokens = input_tokens
            self.time_to_first_token = prompt_time_ms / 1000.0
            return

        # Parse generation evaluation line
        eval_match = re.search(
            r"eval time\s*=\s*([\d.]+)\s*ms\s*/\s*(\d+)\s*tokens.*?"
            r"([\d.]+)\s*tokens per second",
            line,
        )
        if eval_match:
            eval_time_ms = float(eval_match.group(1))
            output_tokens = int(eval_match.group(2))
            tokens_per_second = float(eval_match.group(3))

            self.eval_time = eval_time_ms / 1000.0
            self.output_tokens = output_tokens
            self.tokens_per_second = tokens_per_second
            return

    def get_telemetry_data(self):
        return {
            "input_tokens": self.input_tokens,
            "output_tokens": self.output_tokens,
            "time_to_first_token": self.time_to_first_token,
            "tokens_per_second": self.tokens_per_second,
            "decode_token_times": None,
        }

    def show_telemetry(self):
        # Check if debug logging is enabled
        if not logging.getLogger().isEnabledFor(logging.DEBUG):
            return

        # Prepare telemetry data (transposed format)
        telemetry = [
            ["Input tokens", self.input_tokens],
            ["Output tokens", self.output_tokens],
            ["TTFT (s)", f"{self.time_to_first_token:.2f}"],
            ["TPS", f"{self.tokens_per_second:.2f}"],
        ]

        table = tabulate(
            telemetry, headers=["Metric", "Value"], tablefmt="fancy_grid"
        ).split("\n")

        # Show telemetry in debug while complying with uvicorn's log indentation
        logging.debug("\n          ".join(table))


def _log_subprocess_output(
    process: subprocess.Popen, prefix: str, telemetry: LlamaTelemetry
):
    """
    Read subprocess output line by line, log to debug, and parse telemetry
    """

    if process.stdout:
        try:
            for line in iter(process.stdout.readline, ""):
                if line:
                    line_stripped = line.strip()
                    logging.debug("%s: %s", prefix, line_stripped)

                    telemetry.parse_telemetry_line(line_stripped)

                if process.poll() is not None:
                    break
        except UnicodeDecodeError as e:
            logging.debug("Unicode decode error reading subprocess output: %s", str(e))
        except Exception as e:  # pylint: disable=broad-exception-caught
            logging.error("Unexpected error reading subprocess output: %s", str(e))


def _wait_for_load(llama_server_process: subprocess.Popen, port: int):
    status_code = None
    while not llama_server_process.poll() and status_code != 200:
        health_url = f"http://localhost:{port}/health"
        try:
            health_response = requests.get(health_url)
        except requests.exceptions.ConnectionError:
            logging.debug("Not able to connect to llama-server yet, will retry")
        else:
            status_code = health_response.status_code
            logging.debug(
                "Testing llama-server readiness (will retry until ready), "
                f"result: {health_response.json()}"
            )
        time.sleep(1)


def _launch_llama_subprocess(
    snapshot_files: dict,
    use_gpu: bool,
    telemetry: LlamaTelemetry,
    supports_embeddings: bool = False,
    supports_reranking: bool = False,
) -> subprocess.Popen:
    """
    Launch llama server subprocess with appropriate configuration.

    Args:
        snapshot_files: Dictionary of model files to load
        use_gpu: Whether to use GPU acceleration
        telemetry: Telemetry object for tracking performance metrics
        supports_embeddings: Whether the model supports embeddings
        supports_reranking: Whether the model supports reranking

    Returns:
        Subprocess handle for the llama server
    """

    # Get the current executable path (handles both Windows and Ubuntu structures)
    _, exe_path = get_llama_server_paths()

    # Build the base command
    base_command = [exe_path, "-m", snapshot_files["variant"]]
    if "mmproj" in snapshot_files:
        base_command.extend(["--mmproj", snapshot_files["mmproj"]])
        if not use_gpu:
            base_command.extend(["--no-mmproj-offload"])

    # Find a port, and save it in the telemetry object for future reference
    # by other functions
    telemetry.choose_port()

    # Add port and jinja to enable tool use
    base_command.extend(["--port", str(telemetry.port), "--jinja"])

    # Use legacy reasoning formatting, since not all apps support the new
    # reasoning_content field
    base_command.extend(["--reasoning-format", "none"])

    # Add embeddings support if the model supports it
    if supports_embeddings:
        base_command.append("--embeddings")

    # Add reranking support if the model supports it
    if supports_reranking:
        base_command.append("--reranking")

    # Configure GPU layers: 99 for GPU, 0 for CPU-only
    ngl_value = "99" if use_gpu else "0"
    command = base_command + ["-ngl", ngl_value]

    # Set up environment with library path for Linux
    env = os.environ.copy()
    if platform.system().lower() == "linux":
        lib_dir = os.path.dirname(exe_path)  # Same directory as the executable
        current_ld_path = env.get("LD_LIBRARY_PATH", "")
        if current_ld_path:
            env["LD_LIBRARY_PATH"] = f"{lib_dir}:{current_ld_path}"
        else:
            env["LD_LIBRARY_PATH"] = lib_dir
        logging.debug(f"Set LD_LIBRARY_PATH to {env['LD_LIBRARY_PATH']}")

    # Start subprocess with output capture
    process = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
        encoding="utf-8",
        errors="replace",
        bufsize=1,
        env=env,
    )

    # Start background thread to log subprocess output
    device_type = "GPU" if use_gpu else "CPU"
    threading.Thread(
        target=_log_subprocess_output,
        args=(process, f"LLAMA SERVER {device_type}", telemetry),
        daemon=True,
    ).start()

    return process


def server_load(model_config: PullConfig, telemetry: LlamaTelemetry):
    # Validate platform support before proceeding
    validate_platform_support()

    # Get platform-specific paths at runtime
    llama_server_exe_dir, llama_server_exe_path = get_llama_server_paths()

    # Check whether the llamacpp install needs an upgrade
    version_txt_path = os.path.join(llama_server_exe_dir, "version.txt")
    if os.path.exists(version_txt_path):
        with open(version_txt_path, "r", encoding="utf-8") as f:
            llamacpp_installed_version = f.read()

        if llamacpp_installed_version != LLAMA_VERSION:
            # Remove the existing install, which will trigger a new install
            # in the next code block
            shutil.rmtree(llama_server_exe_dir)

    # Download llama.cpp server if it isn't already available
    if not os.path.exists(llama_server_exe_dir):
        # Download llama.cpp server zip
        llama_zip_url, filename = get_binary_url_and_filename(LLAMA_VERSION)
        llama_zip_path = os.path.join(os.path.dirname(sys.executable), filename)
        logging.info(f"Downloading llama.cpp server from {llama_zip_url}")

        with requests.get(llama_zip_url, stream=True) as r:
            r.raise_for_status()
            with open(llama_zip_path, "wb") as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)

        # Extract zip
        logging.info(f"Extracting {llama_zip_path} to {llama_server_exe_dir}")
        with zipfile.ZipFile(llama_zip_path, "r") as zip_ref:
            zip_ref.extractall(llama_server_exe_dir)

        # Make executable on Linux - need to update paths after extraction
        if platform.system().lower() == "linux":
            # Re-get the paths since extraction might have changed the directory structure
            _, updated_exe_path = get_llama_server_paths()
            if os.path.exists(updated_exe_path):
                os.chmod(updated_exe_path, 0o755)
                logging.info(f"Set executable permissions for {updated_exe_path}")
            else:
                logging.warning(
                    f"Could not find llama-server executable at {updated_exe_path}"
                )

        # Save version.txt
        with open(version_txt_path, "w", encoding="utf-8") as vf:
            vf.write(LLAMA_VERSION)

        # Delete zip file
        os.remove(llama_zip_path)
        logging.info("Cleaned up zip file")

    # Download the gguf to the hugging face cache
    model_manager = ModelManager()
    snapshot_files = model_manager.download_gguf(model_config)
    logging.debug(f"GGUF file paths: {snapshot_files}")

    # Check if model supports embeddings
    supported_models = model_manager.supported_models
    model_info = supported_models.get(model_config.model_name, {})
    supports_embeddings = "embeddings" in model_info.get("labels", [])
    supports_reranking = "reranking" in model_info.get("labels", [])

    # Start the llama-serve.exe process
    logging.debug(f"Using llama_server for GGUF model: {llama_server_exe_path}")

    # Attempt loading on GPU first
    llama_server_process = _launch_llama_subprocess(
        snapshot_files,
        use_gpu=True,
        telemetry=telemetry,
        supports_embeddings=supports_embeddings,
        supports_reranking=supports_reranking,
    )

    # Check the /health endpoint until GPU server is ready
    _wait_for_load(
        llama_server_process,
        telemetry.port,
    )

    # If loading on GPU failed, try loading on CPU
    if llama_server_process.poll():
        logging.warning(
            f"Loading {model_config.model_name} on GPU didn't work, re-attempting on CPU"
        )

        if os.environ.get("LEMONADE_LLAMACPP_NO_FALLBACK"):
            # Used for testing, when the test should fail if GPU didn't work
            raise Exception("llamacpp GPU loading failed")

        llama_server_process = _launch_llama_subprocess(
            snapshot_files,
            use_gpu=False,
            telemetry=telemetry,
            supports_embeddings=supports_embeddings,
            supports_reranking=supports_reranking,
        )

        # Check the /health endpoint until CPU server is ready
        _wait_for_load(
            llama_server_process,
            telemetry.port,
        )

    if llama_server_process.poll():
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"Failed to load {model_config.model_name} with llama.cpp",
        )

    return llama_server_process


def chat_completion(
    chat_completion_request: ChatCompletionRequest, telemetry: LlamaTelemetry
):
    base_url = llamacpp_address(telemetry.port)
    client = OpenAI(
        base_url=base_url,
        api_key="lemonade",
    )

    # Convert Pydantic model to dict and remove unset/null values
    request_dict = chat_completion_request.model_dump(
        exclude_unset=True, exclude_none=True
    )

    # Check if streaming is requested
    if chat_completion_request.stream:

        def event_stream():
            try:
                # Enable streaming
                for chunk in client.chat.completions.create(**request_dict):
                    yield f"data: {chunk.model_dump_json()}\n\n"
                yield "data: [DONE]\n\n"

                # Show telemetry after completion
                telemetry.show_telemetry()

            except Exception as e:  # pylint: disable=broad-exception-caught
                yield f'data: {{"error": "{str(e)}"}}\n\n'

        return StreamingResponse(
            event_stream(),
            media_type="text/event-stream",
            headers={
                "Cache-Control": "no-cache",
                "Connection": "keep-alive",
            },
        )
    else:
        # Non-streaming response
        try:
            # Disable streaming for non-streaming requests
            response = client.chat.completions.create(**request_dict)

            # Show telemetry after completion
            telemetry.show_telemetry()

            return response

        except Exception as e:  # pylint: disable=broad-exception-caught
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail=f"Chat completion error: {str(e)}",
            )


def embeddings(embeddings_request: EmbeddingsRequest, telemetry: LlamaTelemetry):
    """
    Generate embeddings using the llamacpp server.

    Args:
        embeddings_request: The embeddings request containing input text/tokens
        telemetry: Telemetry object containing the server port

    Returns:
        Embeddings response from the llamacpp server
    """
    base_url = llamacpp_address(telemetry.port)
    client = OpenAI(
        base_url=base_url,
        api_key="lemonade",
    )

    # Convert Pydantic model to dict and remove unset/null values
    request_dict = embeddings_request.model_dump(exclude_unset=True, exclude_none=True)

    try:
        # Call the embeddings endpoint
        response = client.embeddings.create(**request_dict)
        return response

    except Exception as e:  # pylint: disable=broad-exception-caught
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"Embeddings error: {str(e)}",
        )


def reranking(reranking_request: RerankingRequest, telemetry: LlamaTelemetry):
    """
    Rerank documents based on their relevance to a query using the llamacpp server.

    Args:
        reranking_request: The reranking request containing query and documents
        telemetry: Telemetry object containing the server port

    Returns:
        Reranking response from the llamacpp server containing ranked documents and scores
    """
    base_url = llamacpp_address(telemetry.port)

    try:
        # Convert Pydantic model to dict and exclude unset/null values
        request_dict = reranking_request.model_dump(
            exclude_unset=True, exclude_none=True
        )

        # Call the reranking endpoint directly since it's not supported by the OpenAI API
        response = requests.post(
            f"{base_url}/rerank",
            json=request_dict,
        )
        response.raise_for_status()
        return response.json()

    except Exception as e:
        logging.error("Error during reranking: %s", str(e))
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"Reranking error: {str(e)}",
        ) from e



from typing import Optional, Union, List, Any

from pydantic import BaseModel

# Set to a high number to allow for interesting experiences in real apps
# Tests should use the max_new_tokens argument to set a lower value
DEFAULT_MAX_NEW_TOKENS = 1500


class LoadConfig(BaseModel):
    """
    Configuration for loading a language model.

    Specifies the model checkpoint, generation parameters,
    and hardware/framework configuration (recipe) for model loading.
    """

    model_name: str
    checkpoint: Optional[str] = None
    recipe: Optional[str] = None
    # Indicates the maximum prompt length allowed for that specific
    # checkpoint + recipe combination
    max_prompt_length: Optional[int] = None
    # Indicates whether the model is a reasoning model, like DeepSeek
    reasoning: Optional[bool] = False
    # Indicates which Multimodal Projector (mmproj) file to use
    mmproj: Optional[str] = None


class CompletionRequest(BaseModel):
    """
    Request model for text completion API endpoint.

    Contains a prompt, a model identifier, and a streaming
    flag to control response delivery.
    """

    prompt: str
    model: str
    echo: bool = False
    stream: bool = False
    logprobs: int | None = False
    stop: list[str] | str | None = None
    temperature: float | None = None
    max_tokens: int | None = None


class ChatCompletionRequest(BaseModel):
    """
    Request model for chat completion API endpoint.

    Contains a list of chat messages, a model identifier,
    and a streaming flag to control response delivery.
    """

    messages: list[dict]
    model: str
    stream: bool = False
    logprobs: int | None = False
    stop: list[str] | str | None = None
    temperature: float | None = None
    tools: list[dict] | None = None
    max_tokens: int | None = None
    max_completion_tokens: int | None = None
    response_format: dict | None = None


class EmbeddingsRequest(BaseModel):
    """
    Request model for embeddings API endpoint.

    Generates embeddings for the provided input text or tokens.
    """

    input: Union[str, List]
    model: Optional[str] = None
    encoding_format: Optional[str] = "float"  # "float" or "base64"


class RerankingRequest(BaseModel):
    """
    Request model for reranking API endpoint.

    Reranks a list of documents based on their relevance to a query.
    """

    query: str
    documents: List[str]
    model: str


class ResponsesRequest(BaseModel):
    """
    Request model for responses API endpoint.
    """

    input: list[dict] | str
    model: str
    max_output_tokens: int | None = None
    temperature: float | None = None
    stream: bool = False


class PullConfig(LoadConfig):
    """
    Pull and load have the same fields.
    """


class DeleteConfig(BaseModel):
    """
    Configuration for deleting a supported LLM.
    """

    model_name: str



import os
import time
import json
import logging
from queue import Queue
from packaging.version import Version
import onnxruntime_genai as og
from transformers import AutoTokenizer
from lemonade.tools.adapter import (
    ModelAdapter,
    TokenizerAdapter,
    PassthroughTokenizerResult,
)


class OrtGenaiTokenizer(TokenizerAdapter):
    def __init__(self, model: og.Model, hf_tokenizer: AutoTokenizer):
        super().__init__(hf_tokenizer)
        # Initialize OGA tokenizer
        self.tokenizer = og.Tokenizer(model)

        # Placeholder value since some code will try to query it
        # If we actually need this to return a proper value, then
        # og.GeneratorParams.eos_token_id has it
        self.eos_token_id = None

    def __call__(self, prompt: str, return_tensors="np"):
        tokens = self.tokenizer.encode(prompt)
        return PassthroughTokenizerResult(tokens)

    # pylint: disable=unused-argument
    def decode(self, response, skip_special_tokens=True) -> str:
        return self.tokenizer.decode(response)


class OrtGenaiStreamer:
    def __init__(self, tokenizer: OrtGenaiTokenizer, timeout=None):
        self.tokenizer = tokenizer
        self.text_queue = Queue()
        self.stop_signal = None
        self.timeout = timeout

    def add_text(self, text: str):
        self.text_queue.put(text, timeout=self.timeout)

    def done(self):
        self.text_queue.put(self.stop_signal, timeout=self.timeout)

    def __iter__(self):
        return self

    def __next__(self):
        value = self.text_queue.get(timeout=self.timeout)
        if value == self.stop_signal:
            raise StopIteration()
        else:
            return value


class OrtGenaiModel(ModelAdapter):

    def __init__(self, input_folder):
        super().__init__()
        self.model = og.Model(input_folder)
        self.type = "ort-genai"
        self.config = self.load_config(input_folder)

    def load_config(self, input_folder):
        rai_config_path = os.path.join(input_folder, "rai_config.json")
        if os.path.exists(rai_config_path):
            with open(rai_config_path, "r", encoding="utf-8") as f:
                max_prompt_length = json.load(f)["max_prompt_length"]["1.4.1"]
        else:
            max_prompt_length = None

        config_path = os.path.join(input_folder, "genai_config.json")
        if os.path.exists(config_path):
            with open(config_path, "r", encoding="utf-8") as f:
                config_dict = json.load(f)
                if max_prompt_length:
                    config_dict["max_prompt_length"] = max_prompt_length
                return config_dict
        return None

    def generate(
        self,
        input_ids,
        max_new_tokens=512,
        min_new_tokens=0,
        do_sample=True,
        top_k=50,
        top_p=1.0,
        temperature=0.7,
        streamer: OrtGenaiStreamer = None,
        pad_token_id=None,
        stopping_criteria=None,
        max_length=None,
        random_seed=1,
    ):
        params = og.GeneratorParams(self.model)

        prompt_length = len(input_ids)
        max_prompt_length = self.config.get("max_prompt_length")
        if max_prompt_length and prompt_length > max_prompt_length:
            raise ValueError(
                f"This prompt (length {prompt_length}) exceeds the model's "
                f"maximum allowed prompt length ({max_prompt_length})."
            )

        # There is a breaking API change in OGA 0.6.0
        # Determine whether we should use the old or new APIs
        # This also supports 0.6.0.dev0, which evaluates to less than 0.6.0 in Version
        use_oga_post_6_api = (
            Version(og.__version__) >= Version("0.6.0") or "0.6.0" in og.__version__
        )
        use_oga_pre_6_api = not use_oga_post_6_api

        if pad_token_id:
            params.pad_token_id = pad_token_id

        # Handle max_length and max_new_tokens
        if max_length and max_new_tokens:
            logging.warning(
                "Both max_length and max_new_tokens were provided. "
                "max_length will take precedence. "
                "When setting max_length, please explicitly set max_new_tokens to None."
            )
        max_length_to_use = None
        if max_length:
            max_length_to_use = max_length
        elif max_new_tokens:
            max_length_to_use = prompt_length + max_new_tokens

        min_length = prompt_length + min_new_tokens

        if use_oga_pre_6_api:
            params.input_ids = input_ids

        if random_seed is None:
            random_seed = -1  # In og.Generator, -1 = seed with random device

        if self.config and "search" in self.config:
            search_config = self.config["search"]
            params.set_search_options(
                do_sample=search_config.get("do_sample", do_sample),
                top_k=search_config.get("top_k", top_k),
                top_p=search_config.get("top_p", top_p),
                temperature=search_config.get("temperature", temperature),
                max_length=max_length_to_use,
                min_length=min_length,
                early_stopping=search_config.get("early_stopping", False),
                length_penalty=search_config.get("length_penalty", 1.0),
                num_beams=search_config.get("num_beams", 1),
                num_return_sequences=search_config.get("num_return_sequences", 1),
                repetition_penalty=search_config.get("repetition_penalty", 1.0),
                past_present_share_buffer=search_config.get(
                    "past_present_share_buffer", True
                ),
                random_seed=random_seed,
                # Not currently supported by OGA
                # diversity_penalty=search_config.get('diversity_penalty', 0.0),
                # no_repeat_ngram_size=search_config.get('no_repeat_ngram_size', 0),
            )
        else:
            params.set_search_options(
                do_sample=do_sample,
                top_k=top_k,
                top_p=top_p,
                temperature=temperature,
                max_length=max_length_to_use,
                min_length=min_length,
                random_seed=random_seed,
            )
        params.try_graph_capture_with_max_batch_size(1)

        generator = og.Generator(self.model, params)

        if streamer is None:
            prompt_start_time = time.perf_counter()
            if use_oga_post_6_api:
                generator.append_tokens(input_ids)
            if use_oga_pre_6_api:
                generator.compute_logits()
            generator.generate_next_token()
            prompt_end_time = time.perf_counter()

            self.time_to_first_token = prompt_end_time - prompt_start_time

            if max_new_tokens > 1:

                token_gen_times = []
                while not generator.is_done():
                    token_gen_start_time = time.perf_counter()
                    if use_oga_pre_6_api:
                        generator.compute_logits()
                    generator.generate_next_token()
                    token_gen_end_time = time.perf_counter()

                    token_gen_times.append(token_gen_end_time - token_gen_start_time)

                if token_gen_times:
                    # List will be empty if we generated 1 or 0 tokens, and we don't
                    # want a divide-by-zero error in those cases
                    avg_token_gen_latency_s = sum(token_gen_times) / len(
                        token_gen_times
                    )
                    self.tokens_per_second = 1 / avg_token_gen_latency_s

            return [generator.get_sequence(0)]
        else:
            if use_oga_post_6_api:
                generator.append_tokens(input_ids)
            tokenizer_stream = streamer.tokenizer.tokenizer.create_stream()

            stop_early = False

            while not generator.is_done() and not stop_early:
                if use_oga_pre_6_api:
                    generator.compute_logits()
                generator.generate_next_token()

                new_token = generator.get_next_tokens()[0]
                new_text = tokenizer_stream.decode(new_token)

                streamer.add_text(new_text)

                if stopping_criteria is not None:
                    if stopping_criteria[0].stop_event.is_set():
                        stop_early = True

            streamer.done()

    def _model_call(self, input_ids):
        """
        Run the model on input_ids and get logits.

        This method directly accesses model logits rather than using the full generate pipeline for
        several important reasons:
        1. Purpose: We need raw logits from a single forward pass, while generate() is optimized for
           producing multiple tokens through iterative inference
        2. Efficiency: Direct access is more efficient for logprob calculations with no
           sampling overhead
        3. Precision: Logprob calculations require exact control over input-to-output mapping
        4. Consistency: Similar approach used in both HF and OGA implementations

        Args:
            input_ids: Input token IDs

        Returns:
            Logits for each token in the sequence
        """
        import torch

        # Setup generator params
        params = og.GeneratorParams(self.model)

        # Configure for a simple forward pass
        params.set_search_options(
            do_sample=False,
            temperature=0.0,
            max_length=len(input_ids),
        )

        # Initialize generator
        generator = og.Generator(self.model, params)

        # Feed tokens to model based on API version
        generator.append_tokens(input_ids)

        # Extract logits - this returns a list of logits tensors
        logits = generator.get_output("logits")

        # Convert to torch tensor for easier processing
        return torch.tensor(logits[0])

    def _select_cont_toks(self, logits, context_len, continuation_tokens):
        """
        Select and process logits for continuation tokens.

        Args:
            logits: Full sequence logits
            context_len: Length of context tokens
            continuation_tokens: List or tensor of continuation token IDs

        Returns:
            Log probabilities for continuation tokens
        """
        import torch

        # Extract relevant logits for continuation prediction (shift by one)
        cont_logits = logits[
            context_len - 1 : context_len - 1 + len(continuation_tokens)
        ]

        # Convert to torch tensors if needed
        if not isinstance(continuation_tokens, torch.Tensor):
            continuation_tokens = torch.tensor(continuation_tokens, dtype=torch.long)

        # Apply log softmax to get log probabilities
        log_probs = torch.log_softmax(cont_logits, dim=-1)

        # Get log probs for the specific continuation tokens
        token_log_probs = torch.gather(
            log_probs, 1, continuation_tokens.unsqueeze(-1)
        ).squeeze(-1)

        return token_log_probs

    def compute_logprobs(
        self, text, tokenizer, prompt_length=None, logprobs=None, echo=False
    ):
        """
        Compute log probabilities for all tokens in the given text.

        Args:
            text: The full text to analyze (e.g., prompt + completion)
            prompt_length: Number of tokens in the prompt. If provided and echo=False,
                only completion tokens after this position will be returned.
            logprobs: If not None, return log probabilities. Value indicates how many top
                alternatives to return. If True but not an integer, defaults to 5 alternatives.
            echo: If True, include logprobs for prompt tokens. If False, only return logprobs
                for completion tokens.

        Returns:
            - text_offset: Character offsets for each token in the text
            - token_logprobs: Log probability for each token
            - tokens: The actual tokens used
            - top_logprobs: Top alternative log probabilities for each position
        """
        import torch

        if tokenizer is None:
            raise ValueError("Tokenizer is required for logprob calculation")

        # Encode the full text
        tokens = tokenizer(text).input_ids  # pylint: disable=E1102

        # Track character offsets for each token
        text_offset = []
        start_idx = 0

        token_strings = []
        for token_id in tokens:
            token_str = tokenizer.decode([token_id])
            token_strings.append(token_str)

            # Calculate character offsets for tokens - handles cases where tokens
            # may not directly match in the original text due to encoding differences,
            # special characters, or tokenization artifacts
            try:
                pos = text[start_idx:].find(token_str)
                if pos != -1:
                    text_offset.append(start_idx + pos)
                    start_idx += pos + len(token_str)
                else:
                    text_offset.append(start_idx)
            except (TypeError, ValueError, UnicodeError):
                # Fallback to current position when matching fails due to encoding issues
                text_offset.append(start_idx)

        # Get logits from model
        logits = self._model_call(tokens)

        # Calculate log probabilities for each token
        all_log_probs = torch.log_softmax(logits, dim=-1)

        # The first token doesn't have a conditional probability
        # For tokens after the first, get the predicted probability
        token_log_probs = []
        top_logprobs_list = []

        # For each position, get the actual token probability and top alternatives
        for i in range(len(tokens)):
            # Get previous token position logits
            if i > 0:  # First token has no preceding context
                prev_logits = all_log_probs[i - 1]
                curr_token_id = tokens[i]
                # Get probability of the actual token that appeared
                token_logprob = prev_logits[curr_token_id].item()
                token_log_probs.append(token_logprob)

                # Get top-k alternatives if requested
                if logprobs is not None:
                    num_alternatives = logprobs if isinstance(logprobs, int) else 5
                    topk_values, topk_indices = torch.topk(
                        prev_logits, min(num_alternatives, prev_logits.size(-1))
                    )

                    # Create dictionary of token: logprob
                    position_logprobs = {}
                    for val, idx in zip(topk_values.tolist(), topk_indices.tolist()):
                        token_str = tokenizer.decode([idx])
                        position_logprobs[token_str] = val

                    top_logprobs_list.append(position_logprobs)
            else:
                # For the first token, we don't have a conditional probability
                token_log_probs.append(None)
                top_logprobs_list.append({})

        # If we don't want to echo prompt tokens, filter them out
        if not echo and prompt_length is not None:
            # Ensure prompt_length is within bounds
            prompt_length = min(prompt_length, len(tokens))

            # Filter results to only include completion tokens
            if prompt_length < len(tokens):
                filtered_text_offset = text_offset[prompt_length:]
                filtered_token_logprobs = token_log_probs[prompt_length:]
                filtered_tokens = token_strings[prompt_length:]
                filtered_top_logprobs = top_logprobs_list[prompt_length:]

                return (
                    filtered_text_offset,
                    filtered_token_logprobs,
                    filtered_tokens,
                    filtered_top_logprobs,
                )
            else:
                # No completion tokens
                return [], [], [], []

        return text_offset, token_log_probs, token_strings, top_logprobs_list



import os
from typing import Optional
import socket
from huggingface_hub import model_info


def is_offline():
    """
    Check if the system is offline by attempting to connect to huggingface.co.

    Returns:
        bool: True if the system is offline (cannot connect to huggingface.co),
              False otherwise.
    """
    if os.environ.get("LEMONADE_OFFLINE"):
        return True
    try:
        socket.gethostbyname("huggingface.co")
        return False
    except socket.gaierror:
        return True


def get_base_model(checkpoint: str) -> Optional[str]:
    """
    Get the base model information for a given checkpoint from the Hugging Face Hub.
    Will auto-detect if we're offline and skip the network call in that case.

    Args:
        checkpoint: The model checkpoint to query

    Returns:
        The base model name if found, or None if not found or error occurs
    """
    # Skip network call in offline mode
    if is_offline():
        return None

    try:
        info = model_info(checkpoint)
        if info.cardData and "base_model" in info.cardData:
            if info.cardData["base_model"] is not None:
                # This is a derived model
                return info.cardData["base_model"]
            else:
                # This is itself a base model
                return [checkpoint]
    except Exception:  # pylint: disable=broad-except
        pass
    return None



import argparse
import os
import sys

from lemonade.state import State
from lemonade.tools import Tool
import lemonade.common.printing as printing
import lemonade.common.build as build
from lemonade_install.install import DEFAULT_QUARK_DIR


class QuarkLoad(Tool):
    """
    Load a model Quantized and exported using Quark.
    Required Input State:
        - state.model: Pretrained model instance to be quantized.
        - state.tokenizer: Tokenizer instance from Hugging Face.
    Output:
        - state of the loaded model

    See docs/dev_cli/quark.md for more details.
    """

    unique_name = "quark-load"

    def __init__(self):
        super().__init__(monitor_message="Load Quark Quantized model")

    @staticmethod
    def parser(add_help: bool = True) -> argparse.ArgumentParser:
        parser = __class__.helpful_parser(
            short_description="Load a quantized model using Quark",
            add_help=add_help,
        )

        parser.add_argument(
            "--quant-scheme",
            type=str,
            required=True,
            default=None,
            help="Supported quantization schemes in Quark",
        )

        parser.add_argument(
            "--quant-algo",
            type=str,
            required=True,
            default=None,
            choices=["awq", "gptq", "autosmoothquant", None],
            help="Supported quantization algorithms in Quark",
        )

        parser.add_argument(
            "--torch-compile", action="store_true", help="Model torch compile"
        )

        parser.add_argument(
            "--safetensors-model-reload",
            action="store_true",
            help="Safetensors model reload",
        )

        parser.add_argument(
            "--safetensors-model-dir",
            default=None,
            help="Directory of safetensors model",
        )

        parser.add_argument(
            "--params-load", action="store_true", help="Model parameters load"
        )

        parser.add_argument("--json-path", help="Specify the path of saved json file")

        parser.add_argument(
            "--safetensors-path",
            default=None,
            help="Specify the path of saved safetensors file",
        )

        return parser

    def run(
        self,
        state: State,
        quant_scheme: str,
        quant_algo: str,
        torch_compile: bool = False,
        safetensors_model_reload: bool = False,
        safetensors_model_dir: str = None,
        params_load: bool = False,
        json_path: str = None,
        safetensors_path: str = None,
    ) -> State:
        """
        Executes the QuarkLoad process.
        Returns:
            State: The updated state after loading the model.
        Raises:
            Exception: If an error occurs during the QuarkLoad process.
        """

        import torch

        try:
            if os.path.isdir(DEFAULT_QUARK_DIR):
                quark_llm_path = os.path.join(
                    DEFAULT_QUARK_DIR, "examples", "torch", "language_modeling"
                )
                sys.path.insert(0, quark_llm_path)
            else:
                raise FileNotFoundError(
                    f"The directory {DEFAULT_QUARK_DIR} does not exist. \
                        Please check your installation."
                )

            # Default load path specific to recipe
            # This will NOT work
            # The default path is now uniquely craeated with timestamp
            # Default load path will not work. Need to pass explicit load path
            model_export_path = os.path.join(
                build.output_dir(state.cache_dir, state.build_name),
                "exported_model",
                quant_scheme,
                quant_algo,
            )

            # Set default paths only if current values are None
            if safetensors_model_dir is None:
                safetensors_model_dir = model_export_path
            if safetensors_path is None:
                safetensors_path = os.path.join(model_export_path, "model.safetensors")
            printing.log_info("Loading model ...")
            if not params_load and not safetensors_model_reload:
                raise ValueError(
                    " Specify load format: 'params_load' or 'safetensors_model_reload'."
                )

            # Reload quantized model if specified
            from quark.torch import load_params, import_model_info

            if params_load:
                printing.log_info(
                    "Restoring quantized model from JSON/safetensors files"
                )
                model = load_params(
                    model,
                    json_path=json_path,
                    safetensors_path=safetensors_path,
                )
            elif safetensors_model_reload:
                printing.log_info(
                    "Restoring quantized model from quark_safetensors files"
                )
                model = import_model_info(model, model_info_dir=safetensors_model_dir)

            if torch_compile:
                printing.log_info("torch.compile...")
                model = torch.compile(model)

            state.model = model
            state.dtype = model.dtype

            printing.log_info("Quark Load process completed.")

        except Exception as e:
            printing.log_error(f"An error occurred during the QuarkLoad process: {e}")
            raise
        return state


# This file was originally licensed under Apache 2.0. It has been modified.
# Modifications Copyright (c) 2025 AMD



import argparse
import sys
import os
from typing import Tuple, Optional
import psutil
from typing import List


# Error codes for different CLI scenarios
class ExitCodes:
    SUCCESS = 0
    GENERAL_ERROR = 1
    SERVER_ALREADY_RUNNING = 2
    TIMEOUT_STOPPING_SERVER = 3
    ERROR_STOPPING_SERVER = 4


class PullError(Exception):
    """
    The pull command has failed to install an LLM
    """


class DeleteError(Exception):
    """
    The delete command has failed to delete an LLM
    """


def serve(
    port: int,
    log_level: str = None,
    tray: bool = False,
):
    """
    Execute the serve command
    """

    # Check if Lemonade Server is already running
    _, running_port = get_server_info()
    if running_port is not None:
        print(
            (
                f"Lemonade Server is already running on port {running_port}\n"
                "Please stop the existing server before starting a new instance."
            ),
        )
        sys.exit(ExitCodes.SERVER_ALREADY_RUNNING)

    # Otherwise, start the server
    print("Starting Lemonade Server...")
    from lemonade.tools.server.serve import Server, DEFAULT_PORT, DEFAULT_LOG_LEVEL

    server = Server()
    port = port if port is not None else DEFAULT_PORT
    log_level = log_level if log_level is not None else DEFAULT_LOG_LEVEL

    # Hidden environment variable to enable input truncation (experimental feature)
    truncate_inputs = "LEMONADE_TRUNCATE_INPUTS" in os.environ

    server.run(
        port=port,
        log_level=log_level,
        truncate_inputs=truncate_inputs,
        tray=tray,
    )


def stop():
    """
    Stop the Lemonade Server
    """

    # Check if Lemonade Server is running
    running_pid, running_port = get_server_info()
    if running_port is None:
        print(f"Lemonade Server is not running\n")
        return

    # Stop the server
    try:
        process = psutil.Process(running_pid)

        # Get all child processes (including llama-server)
        children = process.children(recursive=True)

        # Terminate the main process first
        process.terminate()

        # Then terminate llama-server child process (known to be stubborn)
        # We avoid killing other child processes, such as the installer
        for child in children:
            if "llama-server" in child.name():
                try:
                    child.terminate()
                except psutil.NoSuchProcess:
                    pass  # Child already terminated

        # Wait for main process
        process.wait(timeout=10)

        # Kill llama-server child process if it didn't terminate gracefully
        for child in children:
            if "llama-server" in child.name():
                try:
                    if child.is_running():
                        child.kill()
                except psutil.NoSuchProcess:
                    pass  # Child already terminated
    except psutil.NoSuchProcess:
        # Process already terminated
        pass
    except psutil.TimeoutExpired:
        print("Timed out waiting for Lemonade Server to stop.")
        sys.exit(ExitCodes.TIMEOUT_STOPPING_SERVER)
    except Exception as e:  # pylint: disable=broad-exception-caught
        print(f"Error stopping Lemonade Server: {e}")
        sys.exit(ExitCodes.ERROR_STOPPING_SERVER)
    print("Lemonade Server stopped successfully.")


def pull(
    model_names: List[str],
    checkpoint: Optional[str] = None,
    recipe: Optional[str] = None,
    reasoning: bool = False,
    mmproj: str = "",
):
    """
    Install an LLM based on its Lemonade Server model name

    If Lemonade Server is running, use the pull endpoint to download the model
    so that the Lemonade Server instance is aware of the pull.

    Otherwise, use ModelManager to install the model.
    """

    server_running, port = status(verbose=False)

    if server_running:
        import requests

        base_url = f"http://localhost:{port}/api/v1"

        for model_name in model_names:
            payload = {"model_name": model_name}

            if checkpoint and recipe:
                # Add the parameters for registering a new model
                payload["checkpoint"] = checkpoint
                payload["recipe"] = recipe

                if reasoning:
                    payload["reasoning"] = reasoning
                if mmproj:
                    payload["mmproj"] = mmproj

            # Install the model
            pull_response = requests.post(f"{base_url}/pull", json=payload)

            if pull_response.status_code != 200:
                raise PullError(
                    f"Failed to install {model_name}. Check the "
                    "Lemonade Server log for more information. A list of supported models "
                    "is provided at "
                    "https://github.com/lemonade-sdk/lemonade/blob/main/docs/server/server_models.md"
                )
    else:
        from lemonade_server.model_manager import ModelManager

        ModelManager().download_models(
            model_names,
            checkpoint=checkpoint,
            recipe=recipe,
            reasoning=reasoning,
            mmproj=mmproj,
        )


def delete(model_names: List[str]):
    """
    Delete an LLM based on its Lemonade Server model name

    If Lemonade Server is running, use the delete endpoint to delete the model
    so that the Lemonade Server instance is aware of the deletion.

    Otherwise, use ModelManager to delete the model.
    """

    server_running, port = status(verbose=False)

    if server_running:
        import requests

        base_url = f"http://localhost:{port}/api/v1"

        for model_name in model_names:
            # Delete the model
            delete_response = requests.post(
                f"{base_url}/delete", json={"model_name": model_name}
            )

            if delete_response.status_code != 200:
                raise DeleteError(
                    f"Failed to delete {model_name}. Check the "
                    "Lemonade Server log for more information."
                )
    else:
        from lemonade_server.model_manager import ModelManager

        for model_name in model_names:
            ModelManager().delete_model(model_name)


def version():
    """
    Print the version number
    """
    from lemonade import __version__ as version_number

    print(f"{version_number}")


def status(verbose: bool = True) -> Tuple[bool, int]:
    """
    Print the status of the server

    Returns a tuple of:
    1. Whether the server is running
    2. What port the server is running on (None if server is not running)
    """
    _, port = get_server_info()
    if port is None:
        if verbose:
            print("Server is not running")
        return False, None
    else:
        if verbose:
            print(f"Server is running on port {port}")
        return True, port


def is_lemonade_server(pid):
    """
    Check whether or not a given PID corresponds to a Lemonade server
    """
    try:
        process = psutil.Process(pid)

        while True:
            process_name = process.name()
            if process_name in [  # Windows
                "lemonade-server-dev.exe",
                "lemonade-server.exe",
            ] or process_name in [  # Linux
                "lemonade-server-dev",
                "lemonade-server",
            ]:
                return True
            elif "llama-server" in process_name:
                return False
            if not process.parent():
                return False
            process = process.parent()
    except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
        return False
    return False


def get_server_info() -> Tuple[int | None, int | None]:
    """
    Returns a tuple of:
    1. Lemonade Server's PID
    2. The port that Lemonade Server is running on
    """

    # Get all network connections and filter for localhost IPv4 listening ports
    try:
        connections = psutil.net_connections(kind="tcp4")

        for conn in connections:
            if (
                conn.status == "LISTEN"
                and conn.laddr
                and conn.laddr.ip in ["127.0.0.1"]
                and conn.pid is not None
            ):
                if is_lemonade_server(conn.pid):
                    return conn.pid, conn.laddr.port

    except Exception:
        pass

    return None, None


def main():
    parser = argparse.ArgumentParser(
        description="Serve LLMs on CPU, GPU, and NPU.",
        usage=argparse.SUPPRESS,
    )

    # Add version flag
    parser.add_argument(
        "-v", "--version", action="store_true", help="Show version number"
    )

    # Create subparsers for commands
    subparsers = parser.add_subparsers(
        title="Available Commands", dest="command", metavar=""
    )

    # Serve command
    serve_parser = subparsers.add_parser("serve", help="Start server")
    serve_parser.add_argument("--port", type=int, help="Port number to serve on")
    serve_parser.add_argument(
        "--log-level",
        type=str,
        help="Log level for the server",
        choices=["critical", "error", "warning", "info", "debug", "trace"],
        default="info",
    )
    if os.name == "nt":
        serve_parser.add_argument(
            "--no-tray",
            action="store_true",
            help="Do not show a tray icon when the server is running",
        )

    # Status command
    status_parser = subparsers.add_parser("status", help="Check if server is running")

    # Stop command
    stop_parser = subparsers.add_parser("stop", help="Stop the server")

    # Pull command
    pull_parser = subparsers.add_parser(
        "pull",
        help="Install an LLM",
        epilog=(
            "More information: "
            "https://github.com/lemonade-sdk/lemonade/blob/main/docs/server/server_models.md"
        ),
    )
    pull_parser.add_argument(
        "model",
        help="Lemonade Server model name",
        nargs="+",
    )
    pull_parser.add_argument(
        "--checkpoint",
        help="For registering a new model: Hugging Face checkpoint to source the model from",
    )
    pull_parser.add_argument(
        "--recipe",
        help="For registering a new model: lemonade.api recipe to use with the model",
    )
    pull_parser.add_argument(
        "--reasoning",
        help="For registering a new model: whether the model is a reasoning model or not",
        type=bool,
        default=False,
    )
    pull_parser.add_argument(
        "--mmproj",
        help="For registering a new multimodal model: full file name of the .mmproj file in the checkpoint",
    )

    # Delete command
    delete_parser = subparsers.add_parser(
        "delete",
        help="Delete an LLM",
        epilog=(
            "More information: "
            "https://github.com/lemonade-sdk/lemonade/blob/main/docs/server/server_models.md"
        ),
    )
    delete_parser.add_argument(
        "model",
        help="Lemonade Server model name",
        nargs="+",
    )

    args = parser.parse_args()

    if os.name != "nt":
        args.no_tray = True

    if args.version:
        version()
    elif args.command == "serve":
        serve(
            port=args.port,
            log_level=args.log_level,
            tray=not args.no_tray,
        )
    elif args.command == "status":
        status()
    elif args.command == "pull":
        pull(
            args.model,
            checkpoint=args.checkpoint,
            recipe=args.recipe,
            reasoning=args.reasoning,
            mmproj=args.mmproj,
        )
    elif args.command == "delete":
        delete(args.model)
    elif args.command == "stop":
        stop()
    elif args.command == "help" or not args.command:
        parser.print_help()


if __name__ == "__main__":
    main()

# This file was originally licensed under Apache 2.0. It has been modified.
# Modifications Copyright (c) 2025 AMD



# Utility that helps users install software. It is structured like a
# ManagementTool, however it is not a ManagementTool because it cannot
# import any lemonade modules in order to avoid any installation
# collisions on imported modules.
#
# This tool can install Ryzen AI software artifacts (libraries, wheels, etc.).
# The Ryzen AI hybrid artifacts directory hierarchy is:
#
#     RYZEN_AI\hybrid
#         libs
#         wheels
#
# The Ryzen AI npu artifacts directory hierarchy is:
#
#     RYZEN_AI\npu
#         libs
#         wheels
#
# In the above, RYZEN_AI is the path to the folder ryzen_ai that is contained in the
# same folder as the executable python.exe.
# The folder RYZEN_AI also contains a file `version_info.json` that contains information
# about the installed Ryzen AI artifacts.
#
# In any python environment, only one set of artifacts can be installed at a time.
# Python environments created by Lemonade v6.1.x or earlier will need to be recreated.
#
# The Ryzen AI 1.3.0 artifact files use a different directory hierarchy.
# The Ryzen AI 1.3.0 hybrid artifacts directory hierarchy is:
#
#     RYZEN_AI\hybrid\hybrid-llm-artifacts_1.3.0_lounge\hybrid-llm-artifacts\
#         onnxruntime_genai\lib
#         onnxruntime_genai\wheel
#         onnx_utils\bin
#         eula\eula
#
# The Ryzen AI 1.3.0 npu artifacts directory hierarchy is:
#
#     RYZEN_AI\npu\amd_oga\
#         bins\xclbin\stx
#         libs
#         wheels
#

import argparse
import glob
import json
import os
from pathlib import Path
import re
import shutil
import subprocess
import sys
from typing import Optional
import zipfile

DEFAULT_RYZEN_AI_VERSION = "1.4.0"
version_info_filename = "version_info.json"

lemonade_install_dir = Path(__file__).parent.parent.parent
DEFAULT_QUARK_VERSION = "quark-0.6.0"
DEFAULT_QUARK_DIR = os.path.join(
    lemonade_install_dir, "install", "quark", DEFAULT_QUARK_VERSION
)

# List of supported Ryzen AI processor series (can be extended in the future)
SUPPORTED_RYZEN_AI_SERIES = ["300"]

npu_install_data = {
    "1.3.0": {
        "artifacts_zipfile": "ryzen_ai_13_ga/npu-llm-artifacts_1.3.0.zip",
        "license_file": (
            "https://account.amd.com/content/dam/account/en/licenses/download/"
            "amd-end-user-license-agreement.pdf"
        ),
        "license_tag": "Beta ",
    },
    "1.4.0": {
        "artifacts_zipfile": (
            "https://www.xilinx.com/bin/public/openDownload?"
            "filename=npu-llm-artifacts_1.4.0_032925.zip"
        ),
        "license_file": (
            "https://account.amd.com/content/dam/account/en/licenses/download/"
            "amd-end-user-license-agreement.pdf"
        ),
        "license_tag": "",
    },
}

hybrid_install_data = {
    "1.3.0": {
        "artifacts_zipfile": (
            "https://www.xilinx.com/bin/public/openDownload?"
            "filename=hybrid-llm-artifacts_1.3.0_012725.zip"
        ),
        "license_file": (
            "https://www.xilinx.com/bin/public/openDownload?"
            "filename=AMD%20End%20User%20License%20Agreement.pdf"
        ),
        "license_tag": "",
    },
    "1.4.0": {
        "artifacts_zipfile": (
            "https://www.xilinx.com/bin/public/openDownload?"
            "filename=hybrid-llm-artifacts_1.4.0_032925.zip"
        ),
        "license_file": (
            "https://account.amd.com/content/dam/account/en/licenses/download/"
            "amd-end-user-license-agreement.pdf"
        ),
        "license_tag": "",
    },
}

model_prep_install_data = {
    "1.4.0": {
        "model_prep_artifacts_zipfile": (
            "https://www.xilinx.com/bin/public/openDownload?"
            "filename=model-prep-artifacts-1.4.0-0331.zip"
        )
    }
}


def get_ryzen_ai_path(check_exists=True):
    python_executable_path = sys.executable
    python_executable_dir = os.path.dirname(python_executable_path)
    ryzen_ai_folder = os.path.join(python_executable_dir, "ryzen_ai")
    if check_exists:
        if not os.path.isdir(ryzen_ai_folder):
            raise RuntimeError(
                "Please use `lemonade-install --ryzenai` to install Ryzen AI into the current "
                "python environment. For more info, see "
                "https://ryzenai.docs.amd.com/en/latest/llm/high_level_python.html."
            )
    return ryzen_ai_folder


def get_ryzen_ai_version_info():
    ryzen_ai_folder = get_ryzen_ai_path()
    version_info_path = os.path.join(ryzen_ai_folder, version_info_filename)
    if not os.path.isfile(version_info_path):
        raise RuntimeError(
            f"The file {version_info_filename} is missing from the Ryzen AI "
            f"folder {ryzen_ai_folder}.  Please run `lemonade-install` to fix.  For more info, see "
            "https://ryzenai.docs.amd.com/en/latest/llm/high_level_python.html."
        )
    with open(version_info_path, encoding="utf-8") as file:
        version_info = json.load(file)
    return version_info


def get_oga_npu_dir():
    version_info = get_ryzen_ai_version_info()
    version = version_info["version"]
    ryzen_ai_folder = get_ryzen_ai_path()
    if "1.3.0" in version:
        npu_dir = os.path.join(ryzen_ai_folder, "npu", "amd_oga")
    else:
        npu_dir = os.path.join(ryzen_ai_folder, "npu")
    if not os.path.isdir(npu_dir):
        raise RuntimeError(
            f"The npu artifacts are missing from the Ryzen AI folder {ryzen_ai_folder}. "
            " Please run `lemonade-install --ryzenai npu` again to fix.  For more info, see"
            "https://ryzenai.docs.amd.com/en/latest/llm/high_level_python.html."
        )
    return npu_dir, version


def get_oga_hybrid_dir():
    version_info = get_ryzen_ai_version_info()
    version = version_info["version"]
    ryzen_ai_folder = get_ryzen_ai_path()
    if "1.3.0" in version:
        hybrid_dir = os.path.join(
            ryzen_ai_folder,
            "hybrid",
            "hybrid-llm-artifacts_1.3.0_lounge",
            "hybrid-llm-artifacts",
        )
    else:
        hybrid_dir = os.path.join(ryzen_ai_folder, "hybrid")
    if not os.path.isdir(hybrid_dir):
        raise RuntimeError(
            f"The hybrid artifacts are missing from the Ryzen AI folder {ryzen_ai_folder}. "
            " Please run `lemonade-install --ryzenai hybrid` again to fix.  For more info, see "
            "https://ryzenai.docs.amd.com/en/latest/llm/high_level_python.html."
        )
    return hybrid_dir, version


def download_lfs_file(token, file, output_filename):
    """Downloads a file from LFS"""
    import requests

    # Set up the headers for the request
    headers = {
        "Authorization": f"token {token}",
        "Accept": "application/vnd.github.v3+json",
    }

    response = requests.get(
        f"https://api.github.com/repos/aigdat/ryzenai-sw-ea/contents/{file}",
        headers=headers,
    )

    # Check if the request was successful
    if response.status_code == 200:
        # Parse the JSON response to get the download URL
        content = response.json()
        download_url = content.get("download_url")

        if download_url:
            # Download the file from the download URL
            file_response = requests.get(download_url)

            # Write the content to a file
            with open(output_filename, "wb") as file:
                file.write(file_response.content)
        else:
            print("Download URL not found in the response.")
    else:
        raise ValueError(
            "Failed to fetch the content from GitHub API. "
            f"Status code: {response.status_code}, Response: {response.json()}"
        )

    if not os.path.isfile(output_filename):
        raise ValueError(f"Error: {output_filename} does not exist.")


def download_file(url: str, output_filename: str, description: str = None):
    import requests

    try:
        response = requests.get(url)
        if response.status_code != 200:
            raise Exception(
                f"Failed to fetch the content from GitHub API. \
                Status code: {response.status_code}, Response: {response.json()}"
            )

        with open(output_filename, "wb") as file:
            file.write(response.content)

        if not os.path.isfile(output_filename):
            raise Exception(f"\nError: Failed to write to {output_filename}")

    except Exception as e:
        raise Exception(f"\nError downloading {description or 'file'}: {str(e)}")


def unzip_file(zip_path, extract_to):
    """Unzips the specified zip file to the given directory."""
    with zipfile.ZipFile(zip_path, "r") as zip_ref:
        zip_ref.extractall(extract_to)


def check_ryzen_ai_processor():
    """
    Checks if the current system has a supported Ryzen AI processor.

    Raises:
        UnsupportedPlatformError: If the processor is not a supported Ryzen AI models.
    """
    if not sys.platform.startswith("win"):
        raise UnsupportedPlatformError(
            "Ryzen AI installation is only supported on Windows."
        )

    skip_check = os.getenv("RYZENAI_SKIP_PROCESSOR_CHECK", "").lower() in {
        "1",
        "true",
        "yes",
    }
    if skip_check:
        print("[WARNING]: Processor check skipped.")
        return

    is_supported = False
    cpu_name = ""

    try:
        # Use PowerShell command to get processor name
        powershell_cmd = [
            "powershell",
            "-ExecutionPolicy",
            "Bypass",
            "-Command",
            "Get-WmiObject -Class Win32_Processor | Select-Object -ExpandProperty Name",
        ]

        result = subprocess.run(
            powershell_cmd,
            capture_output=True,
            text=True,
            check=True,
        )

        # Extract the CPU name from PowerShell output
        cpu_name = result.stdout.strip()
        if not cpu_name:
            print(
                "[WARNING]: Could not detect processor name. Proceeding with installation."
            )
            return

        # Check for any supported series
        for series in SUPPORTED_RYZEN_AI_SERIES:
            # Look for the series number pattern - matches any processor in the supported series
            pattern = rf"ryzen ai.*\b{series[0]}\d{{2}}\b"
            match = re.search(pattern, cpu_name.lower(), re.IGNORECASE)

            if match:
                is_supported = True
                break

        if not is_supported:
            print(
                f"[WARNING]: Processor '{cpu_name}' may not be officially supported for Ryzen AI hybrid execution."
            )
            print(
                "[WARNING]: Installation will proceed, but hybrid features may not work correctly."
            )
            print("[WARNING]: Officially supported processors: Ryzen AI 300-series")

    except Exception as e:  # pylint: disable=broad-exception-caught
        print(
            f"[WARNING]: Could not detect processor ({e}). Proceeding with installation."
        )
        print("[WARNING]: Hybrid features may not work if processor is not supported.")


def download_and_extract_package(
    url: str,
    version: str,
    install_dir: str,
    package_name: str,
) -> str:
    """
    Downloads, Extracts and Renames the folder

    Args:
        url: Download URL for the package
        version: Version string
        install_dir: Directory to install to
        package_name: Name of the package

    Returns:
        str: Path where package was extracted (renamed to package-version)
    """
    import requests

    zip_filename = f"{package_name}-{version}.zip"
    zip_path = os.path.join(install_dir, zip_filename)
    target_folder = os.path.join(install_dir, f"{package_name}-{version}")

    print(f"\nDownloading {package_name} from {url}")
    response = requests.get(url)
    if response.status_code == 200:
        with open(zip_path, "wb") as f:
            f.write(response.content)
    else:
        raise Exception(
            f"Failed to download {package_name}. Status code: {response.status_code}"
        )

    print("\n[INFO]: Extracting zip file ...")
    with zipfile.ZipFile(zip_path, "r") as zip_ref:
        zip_ref.extractall(install_dir)
    print("\n[INFO]: Extraction completed.")

    os.remove(zip_path)

    extracted_folder = None
    for folder in os.listdir(install_dir):
        folder_path = os.path.join(install_dir, folder)
        if os.path.isdir(folder_path) and folder.startswith(f"{package_name}-"):
            extracted_folder = folder_path
            break

    if extracted_folder is None:
        raise ValueError(
            f"Error: Extracted folder for {package_name} version {version} not found."
        )

    # Rename extracted folder to package-version
    if extracted_folder != target_folder:
        if os.path.exists(target_folder):
            shutil.rmtree(target_folder)  # Remove if already exists
        os.rename(extracted_folder, target_folder)
        print(f"\n[INFO]: Renamed folder to {target_folder}")

    return target_folder


class LicenseRejected(Exception):
    """
    Raise an exception if the user rejects the license prompt.
    """


class UnsupportedPlatformError(Exception):
    """
    Raise an exception if the hardware is not supported.
    """


class Install:
    """
    Installs the necessary software for specific lemonade features.
    """

    @staticmethod
    def parser() -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(
            description="Installs the necessary software for specific lemonade features",
        )

        parser.add_argument(
            "--ryzenai",
            help="Install Ryzen AI software for LLMs. Requires an authentication token. "
            "The 'npu' and 'hybrid' choices install the default "
            f"{DEFAULT_RYZEN_AI_VERSION} version.",
            choices=[
                "npu",
                "hybrid",
                "unified",
                "npu-1.3.0",
                "hybrid-1.3.0",
                "npu-1.4.0",
                "hybrid-1.4.0",
                "unified-1.4.0",
            ],
        )

        parser.add_argument(
            "--build-model",
            action="store_true",
            help="If specified, installs additional model prep artifacts for Ryzen AI.",
        )

        parser.add_argument(
            "-y",
            "--yes",
            action="store_true",
            help="Answer 'yes' to all questions. "
            "Make sure to review all legal agreements before selecting this option.",
        )

        parser.add_argument(
            "--token",
            help="Some software requires an authentication token to download. "
            "If this argument is not provided, the token can come from an environment "
            "variable (e.g., Ryzen AI uses environment variable OGA_TOKEN).",
        )

        parser.add_argument(
            "--quark",
            help="Install Quark Quantization tool for LLMs",
            choices=["0.6.0"],
        )

        return parser

    @staticmethod
    def _get_license_acceptance(version, license_file, license_tag, yes):
        if yes:
            print(
                f"\nYou have accepted the AMD {license_tag}Software End User License "
                f"Agreement for Ryzen AI {version} by providing the `--yes` option. "
                "The license file is available for your review at "
                f"{license_file}\n"
            )
        else:
            print(
                f"\nYou must accept the AMD {license_tag}Software End User License "
                "Agreement in order to install this software. To continue, type the word "
                "yes to assert that you agree and are authorized to agree "
                "on behalf of your organization, to the terms and "
                f"conditions, in the {license_tag}Software End User License Agreement, "
                "which terms and conditions may be reviewed, downloaded and "
                "printed from this link: "
                f"{license_file}\n"
            )

            response = input("Would you like to accept the license (yes/No)? ")
            if response.lower() == "yes" or response.lower() == "y":
                pass
            else:
                raise LicenseRejected("Exiting because the license was not accepted.")

    @staticmethod
    def _install_artifacts(version, install_dir, token, file, wheels_full_path):
        archive_file_name = "artifacts.zip"
        archive_file_path = os.path.join(install_dir, archive_file_name)

        if token:
            token_to_use = token
        else:
            token_to_use = os.environ.get("OGA_TOKEN")

        # Retrieve the installation artifacts
        if os.path.exists(install_dir):
            # Remove any artifacts from a previous installation attempt
            shutil.rmtree(install_dir)
        os.makedirs(install_dir)
        if any(proto in file for proto in ["https:", "http:"]):
            print(f"\nDownloading {file}\n")
            download_file(file, archive_file_path)
        elif "file:" in file:
            local_file = file.replace("file://", "C:/")
            print(f"\nCopying {local_file}\n")
            shutil.copy(local_file, archive_file_path)
        else:
            print(f"\nDownloading {file} from GitHub LFS to {install_dir}\n")
            download_lfs_file(token_to_use, file, archive_file_path)

        # Unzip the file
        print(f"\nUnzipping archive {archive_file_path}\n")
        unzip_file(archive_file_path, install_dir)
        print(f"\nDLLs installed\n")

        # Install all whl files in the specified wheels folder
        if wheels_full_path is not None:
            print(f"\nInstalling wheels from {wheels_full_path}\n")
            if version == "1.3.0":
                # Install one wheel file at a time (1.3.0 npu build only works this way)
                for file in os.listdir(wheels_full_path):
                    if file.endswith(".whl"):
                        install_cmd = (
                            f"{sys.executable} -m pip install "
                            f"{os.path.join(wheels_full_path, file)}"
                        )
                        print(f"\nInstalling {file} with command {install_cmd}\n")
                        subprocess.run(install_cmd, check=True, shell=True)
            else:
                # Install all the wheel files together, allowing pip to work out the dependencies
                wheel_files = glob.glob(os.path.join(wheels_full_path, "*.whl"))
                install_cmd = [sys.executable, "-m", "pip", "install"] + wheel_files
                subprocess.run(
                    install_cmd,
                    check=True,
                    shell=True,
                )

        # Delete the zip file
        print(f"\nCleaning up, removing {archive_file_path}\n")
        os.remove(archive_file_path)

    @staticmethod
    def _install_ryzenai_model_artifacts(ryzen_ai_folder, version):
        """
        Installs the Ryzen AI model prep artifacts if the build-model flag is provided.

        Args:
            ryzen_ai_folder (str): Path to the Ryzen AI installation folder.
            version (str): Version of the Ryzen AI artifacts.
        """

        # Check if model prep artifacts are available for the given version
        if version not in model_prep_install_data:
            raise ValueError(
                "Model prep artifacts are only available "
                "for version 1.4.0 and above, "
                "and only for 'hybrid' and 'npu-only' targets."
            )

        # Get the model prep artifacts zipfile URL
        file = model_prep_install_data[version].get(
            "model_prep_artifacts_zipfile", None
        )
        if not file:
            raise ValueError(
                "Model prep artifacts zipfile URL is missing for version " f"{version}."
            )

        # Download and extract the model prep artifacts
        print(f"\nDownloading model prep artifacts for Ryzen AI {version}.")
        model_prep_dir = download_and_extract_package(
            url=file,
            version=version,
            install_dir=ryzen_ai_folder,
            package_name="model-prep",
        )

        # Install all .whl files from the extracted model prep artifacts
        wheels_full_path = os.path.join(model_prep_dir)
        print(f"\nInstalling model prep wheels from {wheels_full_path}\n")
        wheel_files = glob.glob(os.path.join(wheels_full_path, "*.whl"))
        if not wheel_files:
            raise ValueError(
                f"No .whl files found in the model prep artifacts directory: {wheels_full_path}"
            )

        install_cmd = [sys.executable, "-m", "pip", "install"] + wheel_files
        subprocess.run(install_cmd, check=True, shell=True)

        print(f"\nModel prep artifacts installed successfully from {model_prep_dir}.")
        return file

    @staticmethod
    def _install_ryzenai_npu(ryzen_ai_folder, version, yes, token, skip_wheels=False):
        # Check version is valid
        if version not in npu_install_data:
            raise ValueError(
                "Invalid Ryzen AI version for NPU.  Valid options are "
                f"{list(npu_install_data.keys())}."
            )
        file = npu_install_data[version].get("artifacts_zipfile", None)
        license_file = npu_install_data[version].get("license_file", None)
        license_tag = npu_install_data[version].get("license_tag", None)
        install_dir = os.path.join(ryzen_ai_folder, "npu")
        if version == "1.3.0":
            wheels_full_path = os.path.join(install_dir, "amd_oga/wheels")
        else:
            wheels_full_path = os.path.join(install_dir, "wheels")

        if license_file:
            Install._get_license_acceptance(version, license_file, license_tag, yes)

        print(f"Downloading NPU artifacts for Ryzen AI {version}.")
        if skip_wheels:
            wheels_full_path = None
        else:
            print("Wheels will be added to current activated environment.")
        Install._install_artifacts(version, install_dir, token, file, wheels_full_path)
        return file

    @staticmethod
    def _install_ryzenai_hybrid(
        ryzen_ai_folder, version, yes, token, skip_wheels=False
    ):
        # Check version is valid
        if version not in hybrid_install_data:
            raise ValueError(
                "Invalid version for hybrid.  Valid options are "
                f"{list(hybrid_install_data.keys())}."
            )
        file = hybrid_install_data[version].get("artifacts_zipfile", None)
        license_file = hybrid_install_data[version].get("license_file", None)
        license_tag = hybrid_install_data[version].get("license_tag", None)
        install_dir = os.path.join(ryzen_ai_folder, "hybrid")
        if version == "1.3.0":
            wheels_full_path = os.path.join(
                ryzen_ai_folder,
                "hybrid",
                "hybrid-llm-artifacts_1.3.0_lounge",
                "hybrid-llm-artifacts",
                "onnxruntime_genai",
                "wheel",
            )
        else:
            wheels_full_path = os.path.join(install_dir, "wheels")

        if license_file:
            Install._get_license_acceptance(version, license_file, license_tag, yes)

        print(f"Downloading hybrid artifacts for Ryzen AI {version}.")
        if skip_wheels:
            wheels_full_path = None
        else:
            print("Wheels will be added to current activated environment.")
        Install._install_artifacts(version, install_dir, token, file, wheels_full_path)
        return file

    @staticmethod
    def _install_ryzenai(ryzenai, build_model, yes, token):
        # Check if the processor is supported before proceeding
        check_ryzen_ai_processor()

        # Delete any previous Ryzen AI installation in this environment
        ryzen_ai_folder = get_ryzen_ai_path(check_exists=False)
        if os.path.exists(ryzen_ai_folder):
            print("Deleting previous Ryzen AI installation in this environment.")
            shutil.rmtree(ryzen_ai_folder)

        # Determine Ryzen AI version to install
        version = DEFAULT_RYZEN_AI_VERSION
        if "-" in ryzenai:
            # ryzenai is: (npu|hybrid)-(<VERSION>)
            parts = ryzenai.split("-")
            ryzenai = parts[0]
            version = parts[1]

        # Install artifacts needed for npu, hybrid, or unified (both) inference
        npu_file = None
        hybrid_file = None
        model_prep_file = None
        if ryzenai == "npu":
            npu_file = Install._install_ryzenai_npu(
                ryzen_ai_folder, version, yes, token
            )
        elif ryzenai == "hybrid":
            hybrid_file = Install._install_ryzenai_hybrid(
                ryzen_ai_folder, version, yes, token
            )
        elif ryzenai == "unified":
            # Download NPU artifacts and use the NPU wheels
            npu_file = Install._install_ryzenai_npu(
                ryzen_ai_folder, version, yes, token
            )
            # Download Hybrid artifacts (and skip wheels as they are already loaded)
            hybrid_file = Install._install_ryzenai_hybrid(
                ryzen_ai_folder, version, yes, token, skip_wheels=True
            )
        else:
            raise ValueError(
                f"Value passed to ryzenai argument is not supported: {ryzenai}"
            )

        if build_model:
            model_prep_file = Install._install_ryzenai_model_artifacts(
                ryzen_ai_folder, version
            )

        # Write version_info file
        version_info = {
            "version": version,
            "hybrid_artifacts": hybrid_file,
            "npu_artifacts": npu_file,
            "model_prep_artifacts": model_prep_file,
        }
        version_info_path = os.path.join(ryzen_ai_folder, version_info_filename)
        try:
            with open(version_info_path, "w", encoding="utf-8") as file:
                json.dump(version_info, file, indent=4)
        except IOError as e:
            print(f"An error occurred while writing {version_info_path}: {e}")

    @staticmethod
    def _install_quark(quark):
        quark_install_dir = os.path.join(lemonade_install_dir, "install", "quark")
        os.makedirs(quark_install_dir, exist_ok=True)

        # Install Quark utilities
        quark_url = (
            f"https://www.xilinx.com/bin/public/openDownload?filename=quark-{quark}.zip"
        )
        quark_path = download_and_extract_package(
            url=quark_url,
            version=quark,
            install_dir=quark_install_dir,
            package_name="quark",
        )
        # Install Quark wheel
        wheel_url = (
            "https://www.xilinx.com/bin/public/openDownload?"
            f"filename=quark-{quark}-py3-none-any.whl"
        )
        wheel_path = os.path.join(quark_install_dir, f"quark-{quark}-py3-none-any.whl")
        print(f"\nInstalling Quark wheel from {wheel_url}")
        download_file(wheel_url, wheel_path, "wheel file")

        install_cmd = f"{sys.executable} -m pip install --no-deps {wheel_path}"
        subprocess.run(install_cmd, check=True, shell=True)
        os.remove(wheel_path)

        print(f"\nQuark installed successfully at: {quark_path}")

    def run(
        self,
        ryzenai: Optional[str] = None,
        build_model: Optional[str] = None,
        quark: Optional[str] = None,
        yes: bool = False,
        token: Optional[str] = None,
    ):
        if ryzenai is None and quark is None and models is None:
            raise ValueError(
                "You must select something to install, "
                "for example `--ryzenai`, `--quark`, or `--models`"
            )

        if ryzenai is not None:
            self._install_ryzenai(ryzenai, build_model, yes, token)

        if quark is not None:
            self._install_quark(quark)


def main():
    installer = Install()
    args = installer.parser().parse_args()
    installer.run(**args.__dict__)


if __name__ == "__main__":
    main()

# This file was originally licensed under Apache 2.0. It has been modified.
# Modifications Copyright (c) 2025 AMD



import os
import time
import textwrap
from multiprocessing import Process, Queue
import psutil
import yaml
import lemonade.common.filesystem as fs
import lemonade.common.printing as printing
from lemonade.profilers import Profiler


DEFAULT_TRACK_MEMORY_INTERVAL = 0.25
MEMORY_USAGE_YAML_FILENAME = "memory_usage.yaml"
MEMORY_USAGE_PNG_FILENAME = "memory_usage.png"


class MemoryTracker(Profiler):

    unique_name = "memory"

    @staticmethod
    def add_arguments_to_parser(parser):
        parser.add_argument(
            "-m",
            f"--{MemoryTracker.unique_name}",
            nargs="?",
            metavar="TRACK_INTERVAL",
            type=float,
            default=None,
            const=DEFAULT_TRACK_MEMORY_INTERVAL,
            help="Track memory usage and plot the results. "
            "Optionally, set the tracking interval in seconds "
            f"(default: {DEFAULT_TRACK_MEMORY_INTERVAL})",
        )

    @staticmethod
    def get_time_mem_list(process):
        return [time.time(), process.memory_info().rss]

    def __init__(self, parser_arg_value):
        super().__init__()
        self.status_stats += [fs.Keys.MEMORY_USAGE_PLOT]
        self.track_memory_interval = parser_arg_value
        self.process_being_tracked = None
        self.build_dir = None
        self.queue = None
        self.tracker_process = None
        self.tracking_active = False
        self.yaml_path = None

    def start(self, build_dir):
        if self.tracking_active:
            raise RuntimeError("Cannot start tracking while already tracking")

        # Save the folder where data and plot will be stored
        self.build_dir = build_dir

        # Get the process being tracked
        track_pid = os.getpid()
        self.process_being_tracked = psutil.Process(track_pid)

        # Create queue for passing messages to the tracker
        self.queue = Queue()

        # The yaml file where the memory usage data will be saved
        self.yaml_path = os.path.join(self.build_dir, MEMORY_USAGE_YAML_FILENAME)

        # Create process to continuously sample memory usage
        self.tracker_process = Process(
            target=self._memory_tracker_,
            args=(
                track_pid,
                self.queue,
                self.yaml_path,
                self.track_memory_interval,
            ),
        )
        self.tracker_process.start()
        self.tracking_active = True
        self.set_label("start")
        self.sample()

    def tool_starting(self, tool_name):
        self.set_label(tool_name)

    def tool_stopping(self):
        self.sample()

    def set_label(self, label):
        if self.tracking_active:
            self.queue.put(label)

    def sample(self):
        if self.tracking_active:
            self.queue.put(MemoryTracker.get_time_mem_list(self.process_being_tracked))

    def stop(self):
        if self.tracking_active:
            self.queue.put(None)
            self.tracking_active = False

    def generate_results(self, state, timestamp, _):

        import matplotlib.pyplot as plt

        if self.tracker_process is None:
            return

        if self.tracking_active:
            self.stop()

        # Wait for memory tracker to finish writing yaml data file
        while self.tracker_process.is_alive():
            self.tracker_process.join(timeout=1.0)

        try:
            with open(self.yaml_path, "r", encoding="utf-8") as f:
                memory_tracks = yaml.safe_load(f)
        except FileNotFoundError as e:
            printing.log_info(
                f"Memory tracker file not found: {e.filename}.  No memory usage plot generated"
            )
            state.save_stat(fs.Keys.MEMORY_USAGE_PLOT, None)
            return

        # Add check to ensure that memory_tracks is not empty or improperly formatted
        if not memory_tracks or not isinstance(memory_tracks, list):
            printing.log_info(
                f"Memory tracker file contains no data or is improperly formatted: {self.yaml_path}"
            )
            state.save_stat(fs.Keys.MEMORY_USAGE_PLOT, None)
            return

        # Find final time in the start track (first track) to subtract from all other times
        _, track = memory_tracks[0]
        t0 = track[-1][0]

        # last_t and last_y are used to draw a line between the last point of the prior
        # track and the first point of the current track
        last_t = None
        last_y = None

        plt.figure()
        for k, v in memory_tracks[1:]:
            if len(v) > 0:
                t = [x[0] - t0 for x in v]
                y = [float(x[1]) / 1024**3 for x in v]
                # draw new memory usage track
                if last_t is not None:
                    plt.plot([last_t] + t, [last_y] + y, label=k, marker=".")
                else:
                    plt.plot(t, y, label=k, marker=".")
                last_t = t[-1]
                last_y = y[-1]
        plt.xlabel("Time (sec)")
        plt.ylabel("GB")
        title_str = "Physical Memory Usage\n" + "\n".join(
            textwrap.wrap(state.build_name, 60)
        )
        plt.title(title_str)
        plt.legend()
        plt.grid()
        plt.tight_layout()

        # Save plot to cache and to current folder
        plot_path = os.path.join(
            self.build_dir, f"{timestamp}_{MEMORY_USAGE_PNG_FILENAME}"
        )
        plt.savefig(plot_path)
        plot_path = os.path.join(
            os.getcwd(), f"{timestamp}_{MEMORY_USAGE_PNG_FILENAME}"
        )
        plt.savefig(plot_path)
        state.save_stat(fs.Keys.MEMORY_USAGE_PLOT, plot_path)

    @staticmethod
    def _memory_tracker_(
        tracked_pid,
        input_queue: Queue,
        yaml_path: str,
        track_memory_interval: float,
    ):
        """
        Tracks memory usage during build and saves to yaml file
        The build communicates with the tracker though the input_queue.  It may pass:
          1) string - This is to indicate that a new track is starting and the string is the label
                    for the next segment.  The tracker will automatically track memory usage at
                    the track_memory_interval once a first track_name is given to it.
          2) list - A time and a current memory usage value that is added to the current track
                    (typically used at the end of a segment to make sure that each segment is
                    sampled at least once
          3) None - This indicates that the tracker should stop tracking, save its data to a file
                    and end
        """
        memory_tracks = []
        current_track = []
        track_name = None
        tracker_exit = False

        try:
            tracked_process = psutil.Process(tracked_pid)
            while (
                not tracker_exit and tracked_process.status() == psutil.STATUS_RUNNING
            ):

                time.sleep(track_memory_interval)

                # Read any messages from the parent process
                while not input_queue.empty():
                    try:
                        message = input_queue.get(timeout=0.001)
                        if message is None or isinstance(message, str):
                            # Save current track.
                            if track_name is not None:
                                memory_tracks.append([track_name, current_track])
                            track_name = message
                            current_track = []
                            if message is None:
                                # Wrap up
                                tracker_exit = True
                                break
                        elif isinstance(message, list):
                            # Add time and memory data to current track
                            if track_name is not None:
                                current_track.append(message)
                            else:
                                raise TypeError(
                                    "Track name must be passed to memory tracker prior to "
                                    "sending data"
                                )
                        else:
                            raise TypeError(
                                "Unrecognized message type in memory_tracker input queue: "
                                f"{message}"
                            )

                    except input_queue.Empty:
                        # input_queue.empty had not been updated
                        pass

                if not tracker_exit and track_name is not None:
                    # Save current time and memory usage
                    current_track.append(
                        MemoryTracker.get_time_mem_list(tracked_process)
                    )

            # Save the collected memory tracks
            with open(yaml_path, "w", encoding="utf-8") as f:
                yaml.dump(memory_tracks, f)

        except psutil.NoSuchProcess:
            # If the parent process stopped existing, we can
            # safely assume that tracking is no longer needed
            # NOTE: this only seems to be needed on Windows
            pass


# This file was originally licensed under Apache 2.0. It has been modified.
# Modifications Copyright (c) 2025 AMD



import re
from typing import List, Dict, Pattern, Optional
import logging
import json


def extract_code_block(text: str) -> str:
    """
    Extracts the content inside triple backtick code blocks from a text.

    Args:
        text (str): The text to extract the code block from.

    Returns:
        str: The content of the first code block if any are found, otherwise the raw text.
    """
    # Regex pattern to match triple backtick code blocks (with optional language hint)
    pattern = re.compile(r"```(?:\w+)?\n(.*?)```", re.DOTALL)

    # Find all matches
    code_blocks = pattern.findall(text)

    # Return first match or raw text
    return code_blocks[0] if code_blocks else text


def standardize_tool_call(tool_call: dict) -> dict | None:
    """
    Standardizes the format of tool calls according to the format expected by OpenAI.

    Args:
        tool_call (dict): The tool call to validate.

    Returns:
        dict | None: Standardized tool call if valid, None otherwise.
    """
    # Ensure the tool call has a "name"
    standardized_tool_call = {}
    if "name" in tool_call:
        standardized_tool_call["name"] = tool_call["name"]
    else:
        logging.warning("Tool call does not have a 'name' field.")
        return None

    # Ensure the tool call has "arguments"
    if "arguments" in tool_call:
        standardized_tool_call["arguments"] = tool_call["arguments"]
    elif "parameters" in tool_call:
        standardized_tool_call["arguments"] = tool_call["parameters"]
    else:
        logging.warning("Tool call does not have a 'arguments' or 'parameters' field.")
        return None

    return standardized_tool_call


def get_tool_call_pattern(added_tokens_decoder: List[str]) -> Optional[Pattern]:
    """
    Extracts tool call pattern from the added tokens decoder.
    """
    special_tokens = [v.content for v in added_tokens_decoder.values()]

    # Pattern 1: <tool_call>...</tool_call> block
    # Sample model that uses this pattern: Qwen3-8B
    if "<tool_call>" in special_tokens and "</tool_call>" in special_tokens:
        return re.compile(r"<tool_call>(.*?)</tool_call>", re.DOTALL)

    # Pattern 2: [TOOL_CALLS] [ {...} ] block
    # Sample model that uses this pattern: Mistral-7B-Instruct-v0.3
    elif "[TOOL_CALLS]" in special_tokens:
        return re.compile(r"\[TOOL_CALLS\]\s*\[(.*?)\](?=\s*<|/?eos|$)", re.DOTALL)

    else:
        logging.warning(
            "Tool calling identifiers were not found for the current model."
        )
        return None


def extract_tool_calls(
    text: str, tool_call_pattern: Optional[Pattern] = None
) -> tuple[List[Dict], str]:
    """
    Extracts tool calls from generated text based on tool calling identifiers.

    Args:
        text (str): The text output generated by the model.
        tool_call_pattern (Optional[Pattern]): The pattern to use to extract tool calls.

    Returns:
        tuple[List[Dict], str]: A tuple containing:
            - List[Dict]: A list of extracted tool call objects (raw JSON-like dicts)
            - str: The original text with tool calls removed
    """

    matches = []
    if tool_call_pattern is not None:
        matches = list(tool_call_pattern.finditer(text))

    # Some models don't use any tool calling identifiers.
    # Instead, tool calls are identified by only generating JSON content.
    # Sample model that uses this pattern: Llama-3.1-8B-Instruct
    else:
        try:
            # Remove the json for a code block if needed
            parsed_text = extract_code_block(text)
            json_tool_calls = json.loads(parsed_text)

            if isinstance(json_tool_calls, dict):
                json_tool_calls = [json_tool_calls]

            extracted_tool_calls = []
            for tool_call in json_tool_calls:
                # Return the tool call if all calls are valid
                standard_tool_call = standardize_tool_call(tool_call)
                if standard_tool_call is not None:
                    extracted_tool_calls.append(standard_tool_call)
                else:
                    return [], text

            return extracted_tool_calls, ""

        except json.JSONDecodeError:
            pass

    # Process matches in reverse to avoid position shifting
    extracted_tool_calls = []
    cleaned_text = text
    for match in reversed(matches):
        content = match.group(1).strip()
        json_tool_call = None
        try:
            json_tool_call = json.loads(content)
        except json.JSONDecodeError:
            logging.warning("Could not parse tool call as JSON.")
            continue

        # Attempt to standardize the tool call
        standard_tool_call = standardize_tool_call(json_tool_call)
        if standard_tool_call is None:
            continue

        # If the content is a valid JSON object, add it to the list
        extracted_tool_calls.append(standard_tool_call)

        # Remove the matched tool call from the text
        cleaned_text = cleaned_text[: match.start()] + cleaned_text[match.end() :]

    return extracted_tool_calls, cleaned_text.strip()


# This file was originally licensed under Apache 2.0. It has been modified.
# Modifications Copyright (c) 2025 AMD



import os
import sys

# Add pywin32_system32 dlls (missing from embeddable Python executable)
pywin32_system32 = os.path.join(sys.prefix, "Lib", "site-packages", "pywin32_system32")
os.add_dll_directory(pywin32_system32)

from win32 import win32gui, win32api
import win32.lib.win32con as win32con
import ctypes
import psutil
from typing import Dict, Set, Callable, List, Optional, Tuple, Any

# Windows message constants
WM_USER = 0x0400
WM_TRAYICON = WM_USER + 1
WM_COMMAND = 0x0111


class MenuItem:
    def __init__(
        self,
        text: str,
        callback: Optional[Callable] = None,
        enabled: bool = True,
        submenu=None,
        bitmap_path: Optional[str] = None,
    ):
        self.text = text
        self.callback = callback
        self.enabled = enabled
        self.submenu = submenu
        self.bitmap_path = bitmap_path
        self.id = None  # Will be set when menu is created
        self.bitmap_handle = None


class Menu:
    SEPARATOR = "SEPARATOR"

    def __init__(self, *items):
        self.items = list(items)


class SystemTray:
    """
    Generic system tray implementation for Windows.
    """

    def __init__(self, app_name: str, icon_path: str):
        self.app_name = app_name
        self.icon_path = icon_path
        self.hwnd = None
        self.hinst = win32api.GetModuleHandle(None)
        self.class_atom = None
        self.notify_id = None
        self.menu_handle = None
        self.menu_items = []  # Store menu items with their IDs
        self.next_menu_id = 1000  # Starting ID for menu items

        # Message map for window procedure
        self.message_map = {
            win32con.WM_DESTROY: self.on_destroy,
            win32con.WM_COMMAND: self.on_command,
            WM_TRAYICON: self.on_tray_icon,
        }

    def register_window_class(self):
        """
        Register the window class for the tray icon.
        """
        window_class = win32gui.WNDCLASS()
        window_class.hInstance = self.hinst
        window_class.lpszClassName = f"{self.app_name}TrayClass"
        window_class.style = win32con.CS_VREDRAW | win32con.CS_HREDRAW
        window_class.hCursor = win32api.LoadCursor(0, win32con.IDC_ARROW)
        window_class.hbrBackground = win32con.COLOR_WINDOW
        window_class.lpfnWndProc = self.window_proc

        self.class_atom = win32gui.RegisterClass(window_class)
        return self.class_atom

    def create_window(self):
        """
        Create a hidden window to receive messages.
        """
        style = win32con.WS_OVERLAPPED | win32con.WS_SYSMENU
        self.hwnd = win32gui.CreateWindow(
            self.class_atom,
            f"{self.app_name} Tray",
            style,
            0,
            0,
            win32con.CW_USEDEFAULT,
            win32con.CW_USEDEFAULT,
            0,
            0,
            self.hinst,
            None,
        )

        # Set the AppUserModelID to identify our app to Windows
        try:
            SetCurrentProcessExplicitAppUserModelID = (
                ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID
            )
            SetCurrentProcessExplicitAppUserModelID(ctypes.c_wchar_p(self.app_name))
        except Exception as e:
            print(f"Failed to set AppUserModelID: {e}")

        win32gui.UpdateWindow(self.hwnd)
        return self.hwnd

    def add_tray_icon(self):
        """
        Add the tray icon to the system tray.
        """
        # Load the .ico file
        icon_flags = win32con.LR_LOADFROMFILE | win32con.LR_DEFAULTSIZE
        hicon = win32gui.LoadImage(
            self.hinst, str(self.icon_path), win32con.IMAGE_ICON, 0, 0, icon_flags
        )

        flags = win32gui.NIF_ICON | win32gui.NIF_MESSAGE | win32gui.NIF_TIP
        nid = (self.hwnd, 0, flags, WM_TRAYICON, hicon, self.app_name)
        win32gui.Shell_NotifyIcon(win32gui.NIM_ADD, nid)
        self.notify_id = nid

    def show_balloon_notification(self, title, message, timeout=5000):
        """
        Show a balloon notification from the tray icon.
        """
        if self.notify_id:
            # The notify_id is a tuple, we need to extract its elements
            hwnd, id, flags, callback_msg, hicon, tip = self.notify_id

            # Create a new notification with the balloon info
            flags |= win32gui.NIF_INFO

            # NIIF_USER (0x4) tells Windows to use our custom icon in the notification
            info_flags = 0x4  # NIIF_USER

            # Create the notification data structure with our custom icon flag
            nid = (
                hwnd,
                id,
                flags,
                callback_msg,
                hicon,
                self.app_name,
                message,
                timeout,
                title,
                info_flags,
            )

            # Show the notification
            win32gui.Shell_NotifyIcon(win32gui.NIM_MODIFY, nid)

    def window_proc(self, hwnd, message, wparam, lparam):
        """
        Window procedure to handle window messages.
        """
        if message in self.message_map:
            return self.message_map[message](hwnd, message, wparam, lparam)
        return win32gui.DefWindowProc(hwnd, message, wparam, lparam)

    def on_destroy(self, hwnd, message, wparam, lparam):
        """
        Handle window destruction.
        """
        if self.notify_id:
            win32gui.Shell_NotifyIcon(win32gui.NIM_DELETE, self.notify_id)

        # Clean up bitmap resources
        for item_id, item in self.menu_items:
            if hasattr(item, "bitmap_handle") and item.bitmap_handle:
                win32gui.DeleteObject(item.bitmap_handle)

        win32gui.PostQuitMessage(0)
        return 0

    def on_command(self, hwnd, message, wparam, lparam):
        """
        Handle menu commands.
        """
        menu_id = win32gui.LOWORD(wparam)

        # Find the menu item with this ID and execute its callback
        for item_id, item_obj in self.menu_items:
            if item_id == menu_id and item_obj.callback:
                item_obj.callback(None, item_obj)
                break

        return 0

    def on_tray_icon(self, hwnd, message, wparam, lparam):
        """
        Handle tray icon events.
        """
        if lparam == win32con.WM_RBUTTONUP or lparam == win32con.WM_LBUTTONUP:
            # Show context menu on right-click and left-click
            self.show_menu()
        return 0

    def create_menu_item(self, menu_handle, item, pos):
        """
        Create a menu item and add it to the menu.
        """
        if item == Menu.SEPARATOR:
            win32gui.AppendMenu(menu_handle, win32con.MF_SEPARATOR, 0, "")
            return pos + 1

        # Assign a unique ID to this menu item
        item.id = self.next_menu_id
        self.next_menu_id += 1

        # Store the menu item with its ID for later lookup
        self.menu_items.append((item.id, item))

        # Create the menu item
        flags = win32con.MF_STRING

        if not item.enabled:
            flags |= win32con.MF_GRAYED

        # Add checkmark if this is a selected item
        if hasattr(item, "checked") and item.checked:
            flags |= win32con.MF_CHECKED

        # Handle submenu
        if item.submenu:
            submenu_handle = win32gui.CreatePopupMenu()
            submenu_pos = 0
            for submenu_item in item.submenu.items:
                submenu_pos = self.create_menu_item(
                    submenu_handle, submenu_item, submenu_pos
                )

            # Add submenu to parent menu
            win32gui.AppendMenu(
                menu_handle, win32con.MF_POPUP | flags, submenu_handle, item.text
            )
        else:
            # Add regular menu item first
            win32gui.AppendMenu(menu_handle, flags, item.id, item.text)

            # Then add bitmap if provided (as a separate step)
            if item.bitmap_path:
                try:
                    # Load the bitmap
                    icon_flags = win32con.LR_LOADFROMFILE | win32con.LR_DEFAULTSIZE
                    item.bitmap_handle = win32gui.LoadImage(
                        self.hinst,
                        str(item.bitmap_path),
                        win32con.IMAGE_BITMAP,
                        0,
                        0,
                        icon_flags,
                    )

                    if item.bitmap_handle:
                        # Set the bitmap for checked/unchecked states
                        # Using the same bitmap for both states
                        win32gui.SetMenuItemBitmaps(
                            menu_handle,
                            item.id,
                            win32con.MF_BYCOMMAND,
                            item.bitmap_handle,
                            item.bitmap_handle,
                        )
                    else:
                        print(f"Failed to load bitmap from {item.bitmap_path}")
                except Exception as e:
                    print(f"Error adding bitmap to menu item: {e}")

        return pos + 1

    def build_menu(self, menu_items):
        """
        Build a menu from a list of menu items.
        """
        # Create a new menu
        menu_handle = win32gui.CreatePopupMenu()
        pos = 0

        # Clear previous menu items
        self.menu_items = []
        self.next_menu_id = 1000

        # Add each item to the menu
        for item in menu_items:
            pos = self.create_menu_item(menu_handle, item, pos)

        return menu_handle

    def show_menu(self):
        """
        Show the context menu.
        """
        # Create menu based on current state
        menu = self.create_menu()
        menu_handle = self.build_menu(menu.items)

        # Get cursor position
        pos = win32gui.GetCursorPos()

        # Make our window the foreground window
        try:
            win32gui.SetForegroundWindow(self.hwnd)
        except Exception:
            # Ignore errors when setting foreground window
            # Those are common when the tray icon is already open
            pass

        # Display the menu
        win32gui.TrackPopupMenu(
            menu_handle,
            win32con.TPM_LEFTALIGN | win32con.TPM_RIGHTBUTTON,
            pos[0],
            pos[1],
            0,
            self.hwnd,
            None,
        )

        # Required by Windows
        win32gui.PostMessage(self.hwnd, win32con.WM_NULL, 0, 0)

    def create_menu(self):
        """
        Create the context menu based on current state. Override in subclass.
        """
        return Menu(MenuItem("Exit", self.exit_app))

    def exit_app(self, _, __):
        """Exit the application."""
        win32gui.DestroyWindow(self.hwnd)

    def update_menu(self):
        """
        Update the menu (will be shown on next right-click).
        """
        if self.hwnd:
            win32gui.InvalidateRect(self.hwnd, None, True)

    def message_loop(self):
        """
        Run the Windows message loop.
        """
        win32gui.PumpMessages()

    def run(self):
        """
        Run the tray application.
        """
        # Register window class and create window
        self.register_window_class()
        self.create_window()

        # Add tray icon
        self.add_tray_icon()

        # Run the message loop in the main thread
        self.message_loop()

    def setup_console_control_handler(self, logger=None):
        """
        Set up Windows console control handler for CTRL+C events.
        This handles graceful shutdown when the console window is closed or CTRL+C is pressed.
        """

        def console_ctrl_handler(ctrl_type):
            if ctrl_type in (0, 2):  # CTRL_C_EVENT or CTRL_CLOSE_EVENT
                if logger:
                    logger.info(
                        "Received console control event, shutting down gracefully"
                    )
                # Post a quit message to the main thread instead of calling exit_app directly
                # This avoids thread access issues
                if self.hwnd:
                    win32gui.PostMessage(self.hwnd, win32con.WM_CLOSE, 0, 0)
                return True
            return False

        # Define the handler function type
        HANDLER_ROUTINE = ctypes.WINFUNCTYPE(
            ctypes.wintypes.BOOL, ctypes.wintypes.DWORD
        )
        handler = HANDLER_ROUTINE(console_ctrl_handler)

        # Set the console control handler
        ctypes.windll.kernel32.SetConsoleCtrlHandler(handler, True)

        return handler  # Return handler to keep it in scope



import argparse
from typing import Dict, Optional
import json
from lemonade.tools import FirstTool
from lemonade.state import State
import lemonade.common.status as status
import lemonade.common.printing as printing
from lemonade.cache import Keys


class HuggingfaceLoad(FirstTool):
    """
    Load an LLM as a torch.nn.Module using the Hugging Face transformers
    from_pretrained() API.

    Expected input: a checkpoint to load

    Output state produced:
        - state.model: instance of torch.nn.Module that implements an LLM.
        - state.inputs: tokenized example inputs to the model, in the form of a
            dictionary of kwargs.
        - state.tokenizer: instance of Hugging Face PretrainedTokenizer.
        - state.dtype: data type of the model.
        - state.checkpoint: pretrained checkpoint used to load the model.
    """

    unique_name = "huggingface-load"

    def _imports(self):
        pass

    def __init__(self):
        super().__init__(monitor_message="Loading Huggingface checkpoint")

        self.status_stats = [Keys.DTYPE]

    @staticmethod
    def parser(add_help: bool = True) -> argparse.ArgumentParser:
        parser = __class__.helpful_parser(
            short_description="Load an LLM in PyTorch using huggingface transformers",
            add_help=add_help,
        )

        default_dtype = "float32"
        parser.add_argument(
            "--dtype",
            "-d",
            required=False,
            default=default_dtype,
            help=f"Data type to load the model in (default: {default_dtype}).",
        )

        choices = ["cpu", "cuda"]
        for cuda in range(15):
            choices.append(f"cuda:{cuda}")
        parser.add_argument(
            "--device",
            required=False,
            default=None,
            choices=choices,
            help="Move the model and inputs to a device using the .to() method "
            "(default: don't call the .to() method)",
        )

        parser.add_argument(
            "--load-kwargs",
            required=False,
            default="{}",
            type=json.loads,
            help="Arbitrary kwargs, in json format, that will be passed as "
            "from_pretrained(**kwargs). "
            r"Example: --load-kwargs='{\"trust_remote_code\": true} would result in "
            "from_pretrained(trust_remote_code=True)",
        )

        parser.add_argument(
            "--channels-last",
            default=True,
            type=bool,
            help="Whether to format the model in memory using "
            "channels-last (default: True)",
        )

        return parser

    def parse(self, state: State, args, known_only=True) -> argparse.Namespace:

        from lemonade.tools.huggingface.utils import str_to_dtype

        parsed_args = super().parse(state, args, known_only)

        # Save stats about the user's input (do this prior to decoding)
        state.save_stat(Keys.CHECKPOINT, parsed_args.input)
        state.save_stat(Keys.DTYPE, parsed_args.dtype)

        # Decode dtype arg into a torch value
        parsed_args.dtype = str_to_dtype[parsed_args.dtype]

        return parsed_args

    def run(
        self,
        state: State,
        input: str = "",
        dtype: "torch.dtype" = None,
        device: Optional[str] = None,
        load_kwargs: Optional[Dict] = None,
        channels_last: bool = True,
    ) -> State:
        # Import expensive modules at runtime
        import transformers
        import torch

        from lemonade.tools.huggingface.utils import (
            HuggingfaceTokenizerAdapter,
            HuggingfaceAdapter,
        )
        from lemonade.common.network import (
            is_offline,
            get_base_model,
        )

        # Set default dtype
        if dtype is None:
            dtype_to_use = torch.float32
        else:
            dtype_to_use = dtype

        # Auto-detect offline status
        offline = is_offline()
        if offline:
            printing.log_warning(
                "Network connectivity to huggingface.co not detected. Running in offline mode."
            )

        checkpoint = input

        if load_kwargs is None:
            load_kwargs_to_use = {}
        else:
            load_kwargs_to_use = load_kwargs

        # Add local_files_only to kwargs in offline mode
        if offline:
            load_kwargs_to_use["local_files_only"] = True

        if vars(state).get(Keys.MODEL):
            raise ValueError("HuggingfaceLoad must be the first tool in the sequence")

        try:
            model = transformers.AutoModelForCausalLM.from_pretrained(
                checkpoint,
                torch_dtype=dtype_to_use,
                low_cpu_mem_usage=True,
                **load_kwargs_to_use,
            )
        except Exception as e:
            if offline and "Can't load config for" in str(e):
                raise ValueError(
                    f"Cannot load model {checkpoint} in offline mode. "
                    f"The model files may not be available locally. Original error: {str(e)}"
                )
            raise

        # Only call the model.to() method if an argument to this function
        # provides a reason to do so
        to_args = {}
        if channels_last:
            to_args["memory_format"] = torch.channels_last
        if device:
            to_args["device"] = device
        if to_args:
            model.to(**to_args)

        model = model.eval()

        try:
            tokenizer_kwargs = {
                "use_fast": False,
                "model_max_length": 4096,
                "padding_side": "left",
            }
            if offline:
                tokenizer_kwargs["local_files_only"] = True

            tokenizer = transformers.AutoTokenizer.from_pretrained(
                checkpoint, **tokenizer_kwargs
            )
        except ValueError as e:
            # Sometimes those specific tokenizer flags are not supported, in which
            # case we try to just load a simple tokenizer
            tokenizer_kwargs = {}
            if offline:
                tokenizer_kwargs["local_files_only"] = True

            try:
                tokenizer = transformers.AutoTokenizer.from_pretrained(
                    checkpoint, **tokenizer_kwargs
                )
            except Exception as e:
                if offline and "Can't load tokenizer for" in str(e):
                    raise ValueError(
                        f"Cannot load tokenizer for {checkpoint} in offline mode. "
                        f"The tokenizer files may not be available locally. "
                        f"Original error: {str(e)}"
                    )
                raise

        # Pass the model and inputs into state
        state.model = HuggingfaceAdapter(model, dtype_to_use, device, tokenizer)

        state.tokenizer = HuggingfaceTokenizerAdapter(tokenizer, device)
        state.dtype = dtype_to_use
        state.checkpoint = checkpoint
        state.device = device

        # Save stats about the model
        state.save_stat(Keys.CHECKPOINT, checkpoint)
        state.save_stat(Keys.DTYPE, str(dtype_to_use).split(".")[1])
        state.save_stat(Keys.DEVICE, device)

        # Get base model information
        base_model = get_base_model(checkpoint)
        if base_model is not None:
            state.save_stat("base_model", base_model)

        # Create a UniqueInvocationInfo and ModelInfo so that we can display status
        # at the end of the sequence
        status.add_to_state(state=state, name=input, model=model)

        return state


# This file was originally licensed under Apache 2.0. It has been modified.
# Modifications Copyright (c) 2025 AMD



import os
import sys
import threading
from concurrent.futures import ThreadPoolExecutor
import subprocess
import webbrowser
from pathlib import Path
import logging
import tempfile
import requests
from packaging.version import parse as parse_version

from lemonade.version import __version__
from lemonade.tools.server.utils.system_tray import SystemTray, Menu, MenuItem


class OutputDuplicator:
    """
    Output duplicator that writes to both a file and a stream.
    """

    def __init__(self, file_path, stream):
        self.file = open(file_path, "a", encoding="utf-8")
        self.stream = stream

    def write(self, data):
        self.file.write(data)
        self.stream.write(data)
        self.file.flush()
        self.stream.flush()

    def flush(self):
        self.file.flush()
        self.stream.flush()

    def isatty(self):
        # Delegate the isatty check to the original stream
        return self.stream.isatty()


class LemonadeTray(SystemTray):
    """
    Lemonade-specific system tray implementation.
    """

    def __init__(self, log_file, port, server_factory):
        # Find the icon path
        icon_path = Path(__file__).resolve().parents[0] / "static" / "favicon.ico"

        # Initialize the base class
        super().__init__("Lemonade Server", str(icon_path))

        # Lemonade-specific attributes
        self.loaded_llm = None
        self.server = None
        self.server_thread = None
        self.executor = ThreadPoolExecutor(max_workers=1)
        self.log_file = log_file
        self.port = port
        self.server_factory = server_factory

        # Get current and latest version
        self.current_version = __version__
        self.latest_version = __version__
        self.latest_version_url = None

        # We will get the models list with update_downloaded_models_background()
        # so that getting the list isn't on the critical path of startup
        self.downloaded_models = {}

        # Set up logger
        self.logger = logging.getLogger(__name__)

        # Initialize console handler (will be set up in run method)
        self.console_handler = None

        # Background thread for updating model mapping
        self.model_update_thread = None
        self.stop_model_update = threading.Event()

        # Background thread for version checking
        self.version_check_thread = None
        self.stop_version_check = threading.Event()

    def get_latest_version(self):
        """
        Update the latest version information.
        """
        try:
            response = requests.get(
                "https://api.github.com/repos/lemonade-sdk/lemonade/releases/latest",
                timeout=10,  # Add timeout to prevent hanging
            )
            response.raise_for_status()
            release_data = response.json()

            # Get version from tag name (typically in format "vX.Y.Z")
            self.latest_version = release_data.get("tag_name", "").lstrip("v")

            # Find the installer asset
            self.latest_version_url = None
            for asset in release_data.get("assets", []):
                if asset.get("name", "").endswith(
                    ".exe"
                ) and "Lemonade_Server_Installer" in asset.get("name", ""):
                    self.latest_version_url = asset.get("browser_download_url")
                    break

            self.logger.debug(
                f"Updated version info: latest version {self.latest_version}"
            )
        except Exception as e:  # pylint: disable=broad-exception-caught
            self.logger.error(f"Error fetching latest version: {str(e)}")
        return self.latest_version, self.latest_version_url

    def update_version_background(self):
        """
        Background thread function to update version information periodically.
        """
        self.get_latest_version()
        while not self.stop_version_check.wait(900):  # 900 seconds = 15 minutes
            try:
                self.get_latest_version()
            except Exception as e:  # pylint: disable=broad-exception-caught
                self.logger.error(f"Error updating version in background: {str(e)}")

    def update_downloaded_models_background(self):
        """
        Background thread function to update model mapping every 1 second until a valid
        response is received, then every 10 seconds after that.
        This is used to avoid a ~0.5s delay when opening the tray menu.
        """
        poll_interval = 1
        while not self.stop_model_update.wait(poll_interval):
            try:
                response = requests.get(
                    f"http://localhost:{self.port}/api/v0/models",
                    timeout=0.1,  # Add timeout
                )
                response.raise_for_status()

                # Update the model mapping
                models_endpoint_response = response.json().get("data")

                # Convert to dict
                self.downloaded_models = {
                    model["id"]: {"checkpoint": model["checkpoint"]}
                    for model in models_endpoint_response
                }
                self.logger.debug("Model mapping updated in background")

                # Increase the poling interval after we get our first response
                poll_interval = 10
            except Exception:  # pylint: disable=broad-exception-caught
                # Poll again later
                pass

    def unload_llms(self, _, __):
        """
        Unload the currently loaded LLM.
        """
        requests.post(f"http://localhost:{self.port}/api/v0/unload")

    def load_llm(self, _, __, model_name):
        """Load an LLM model."""

        # Create config for loading
        config = {"model_name": model_name}

        # Use the executor to make the request asynchronously
        self.executor.submit(
            lambda: requests.post(
                f"http://localhost:{self.port}/api/v0/load", json=config
            )
        )

    def show_logs(self, _, __):
        """
        Show the log file in a new window.
        """
        try:
            subprocess.Popen(
                [
                    "powershell",
                    "Start-Process",
                    "powershell",
                    "-ArgumentList",
                    f'"-NoExit", "Get-Content -Wait {self.log_file}"',
                ]
            )
        except Exception as e:  # pylint: disable=broad-exception-caught
            self.logger.error(f"Error opening logs: {str(e)}")

    def open_documentation(self, _, __):
        """
        Open the documentation in the default web browser.
        """
        webbrowser.open("https://lemonade-server.ai/docs/")

    def open_llm_chat(self, _, __):
        """
        Open the LLM chat in the default web browser.
        """
        webbrowser.open(f"http://localhost:{self.port}/#llm-chat")

    def open_model_manager(self, _, __):
        """
        Open the model manager in the default web browser.
        """
        webbrowser.open(f"http://localhost:{self.port}/#model-management")

    def check_server_state(self):
        """
        Check the server state using the health endpoint
        """
        try:
            response = requests.get(
                f"http://localhost:{self.port}/api/v0/health",
                timeout=0.1,  # Add timeout
            )
            response.raise_for_status()
            response_data = response.json()
            checkpoint = response_data.get("model_loaded")

            # Convert checkpoint to model name if possible
            if checkpoint:
                # Create a mapping from checkpoint to model name
                checkpoint_to_model = {
                    model_info["checkpoint"]: model_name
                    for model_name, model_info in self.downloaded_models.items()
                }
                # Use the model name if available, otherwise use the checkpoint
                self.loaded_llm = checkpoint_to_model.get(checkpoint, checkpoint)
            else:
                self.loaded_llm = None

            return True  # Successfully checked server state

        except Exception as e:  # pylint: disable=broad-exception-caught
            # Server might not be ready yet, so just log at debug level
            self.logger.debug(f"Error checking server state: {str(e)}")
            return False  # Failed to check server state

    def change_port(self, _, __, new_port):
        """
        Change the server port and restart the server.
        """
        try:
            # Stop the current server
            if self.server_thread and self.server_thread.is_alive():
                # Set should_exit flag on the uvicorn server instance
                if (
                    hasattr(self.server, "uvicorn_server")
                    and self.server.uvicorn_server
                ):
                    self.server.uvicorn_server.should_exit = True
                self.server_thread.join(timeout=2)

            # Update the port
            self.port = new_port

            # Restart the server
            self.server_thread = threading.Thread(target=self.start_server, daemon=True)
            self.server_thread.start()

            # Show notification
            self.show_balloon_notification(
                "Port Changed", f"Lemonade Server is now running on port {self.port}"
            )

        except Exception as e:  # pylint: disable=broad-exception-caught
            self.logger.error(f"Error changing port: {str(e)}")
            self.show_balloon_notification("Error", f"Failed to change port: {str(e)}")

    def upgrade_to_latest(self, _, __):
        """
        Download and launch the Lemonade Server installer
        """

        # Show notification that download is starting
        self.show_balloon_notification(
            "Upgrading Lemonade",
            "Downloading Lemonade Server Installer. Please wait...",
        )

        # Create temporary file for the installer
        installer_path = os.path.join(
            tempfile.gettempdir(), "Lemonade_Server_Installer.exe"
        )
        if os.path.exists(installer_path):
            os.remove(installer_path)

        # Download the installer
        response = requests.get(self.latest_version_url, stream=True)
        response.raise_for_status()

        # Save the installer to disk and force write to disk
        with open(installer_path, "wb") as f:
            for chunk in response.iter_content(chunk_size=8192):
                f.write(chunk)
            f.flush()
            os.fsync(f.fileno())

        # Launch the installer as a completely detached process
        # subprocess.DETACHED_PROCESS - Creates a process that's not attached to the console
        # subprocess.CREATE_NEW_PROCESS_GROUP - Creates a new process group
        # close_fds=True - Closes file descriptors to prevent inheritance
        subprocess.Popen(
            [installer_path],
            creationflags=subprocess.DETACHED_PROCESS
            | subprocess.CREATE_NEW_PROCESS_GROUP,
            close_fds=True,
            shell=True,
            cwd=tempfile.gettempdir(),
        )

        # No need to quit the application, the installer will handle it

    def create_menu(self):
        """
        Create the Lemonade-specific context menu.
        """
        # Check server health when menu is opened
        status_successfully_checked = self.check_server_state()

        items = []

        if not status_successfully_checked:
            items.append(
                MenuItem("Server Busy - See Logs for details", None, enabled=False)
            )
        elif self.loaded_llm:
            items.extend(
                [
                    MenuItem(f"Loaded: {self.loaded_llm}", None, enabled=False),
                    MenuItem("Unload LLM", self.unload_llms),
                ]
            )
        else:
            items.extend(
                [
                    MenuItem("No models loaded", None, enabled=False),
                ]
            )

        # Create menu items for all downloaded models
        model_menu_items = []
        if not self.downloaded_models:
            model_menu_items.append(
                MenuItem(
                    "No models available: Use the Model Manager to pull models",
                    None,
                    enabled=False,
                )
            )
        else:
            for model_name, _ in self.downloaded_models.items():
                # Create a function that returns the lambda to properly capture the variables
                def create_handler(mod):
                    return lambda icon, item: self.load_llm(icon, item, mod)

                model_item = MenuItem(model_name, create_handler(model_name))

                # Set checked property instead of modifying the text
                model_item.checked = model_name == self.loaded_llm
                model_menu_items.append(model_item)

        load_submenu = Menu(*model_menu_items)

        # Create port selection submenu with 3 options
        port_menu_items = []
        port_options = [
            8000,
            8020,
            8040,
            8060,
            8080,
            9000,
        ]

        for port_option in port_options:
            # Create a function that returns the lambda to properly capture the port variable
            def create_port_handler(port):
                return lambda icon, item: self.change_port(icon, item, port)

            # Set checked property instead of modifying the text
            port_item = MenuItem(
                f"Port {port_option}", create_port_handler(port_option)
            )
            port_item.checked = port_option == self.port
            port_menu_items.append(port_item)

        port_submenu = Menu(*port_menu_items)

        if status_successfully_checked:
            items.append(MenuItem("Load Model", None, submenu=load_submenu))
        items.append(MenuItem("Port", None, submenu=port_submenu))
        items.append(Menu.SEPARATOR)

        # Only show upgrade option if newer version is available
        if parse_version(self.latest_version) > parse_version(self.current_version):
            items.append(
                MenuItem(
                    f"Upgrade to version {self.latest_version}", self.upgrade_to_latest
                )
            )

        items.append(MenuItem("Documentation", self.open_documentation))
        items.append(MenuItem("LLM Chat", self.open_llm_chat))
        items.append(MenuItem("Model Manager", self.open_model_manager))
        items.append(MenuItem("Show Logs", self.show_logs))
        items.append(Menu.SEPARATOR)
        items.append(MenuItem("Quit Lemonade", self.exit_app))
        return Menu(*items)

    def start_server(self):
        """
        Start the uvicorn server.
        """
        self.server = self.server_factory()
        self.server.uvicorn_server = self.server.run_in_thread(port=self.port)
        self.server.uvicorn_server.run()

    def run(self):
        """
        Run the Lemonade tray application.
        """

        # Register window class and create window
        self.register_window_class()
        self.create_window()

        # Set up Windows console control handler for CTRL+C
        self.console_handler = self.setup_console_control_handler(self.logger)

        # Add tray icon
        self.add_tray_icon()

        # Start the background model mapping update thread
        self.model_update_thread = threading.Thread(
            target=self.update_downloaded_models_background, daemon=True
        )
        self.model_update_thread.start()

        # Start the version check thread
        self.version_check_thread = threading.Thread(
            target=self.update_version_background, daemon=True
        )
        self.version_check_thread.start()

        # Start the server in a separate thread
        self.server_thread = threading.Thread(target=self.start_server, daemon=True)
        self.server_thread.start()

        # Show initial notification
        self.show_balloon_notification(
            "Woohoo!",
            (
                "Lemonade Server is running! "
                "Right-click the tray icon below to access options."
            ),
        )

        # Run the message loop in the main thread
        self.message_loop()

    def exit_app(self, icon, item):
        """
        Exit the application.
        """
        # Stop the background threads
        self.stop_model_update.set()
        self.stop_version_check.set()

        if self.model_update_thread and self.model_update_thread.is_alive():
            self.model_update_thread.join(timeout=1)
        if self.version_check_thread and self.version_check_thread.is_alive():
            self.version_check_thread.join(timeout=1)

        # Call parent exit method
        super().exit_app(icon, item)

        # Stop the server using the CLI stop command to ensure a rigorous cleanup
        # This must be a subprocess to ensure the cleanup doesnt kill itself
        subprocess.Popen(
            [sys.executable, "-m", "lemonade_server.cli", "stop"], shell=True
        )



import threading
import logging
from lemonade.tools.server.serve import Server


class ServerRunner(threading.Thread):
    """
    Thread class for running the Lemonade Server with a loaded model.
    """

    def __init__(
        self, model, tokenizer, checkpoint, recipe, host="localhost", port=8000
    ):
        threading.Thread.__init__(self)
        self.model = model
        self.tokenizer = tokenizer
        self.checkpoint = checkpoint
        self.recipe = recipe
        self.host = host
        self.port = port
        self.server = None
        self.ready_event = threading.Event()
        self.shutdown_event = threading.Event()
        self.uvicorn_server = None

    def run(self):
        try:
            # Create the server instance
            self.server = Server()

            # Configure the server with model/tokenizer
            self.server.model = self.model
            self.server.tokenizer = self.tokenizer
            self.server.llm_loaded = type(
                "obj",
                (object,),
                {
                    "checkpoint": self.checkpoint,
                    "recipe": self.recipe,
                    "max_prompt_length": None,
                    "reasoning": False,
                    "model_name": "custom",
                },
            )

            # Set up the server for threaded execution
            self.uvicorn_server = self.server.run_in_thread(
                port=self.port, host=self.host, log_level="warning"
            )

            # Set the ready event
            self.ready_event.set()

            # Run the server until shutdown is requested
            logging.info(f"Starting server on http://{self.host}:{self.port}")
            self.uvicorn_server.run()

        except Exception as e:
            logging.error(f"Error starting server: {e}")
            self.ready_event.set()
            raise

    def shutdown(self):
        """Shutdown the server"""
        if hasattr(self, "uvicorn_server") and self.uvicorn_server:
            logging.info("Shutting down server...")
            self.uvicorn_server.should_exit = True
            self.shutdown_event.set()

        # Clean up resources properly to avoid memory leaks
        if hasattr(self, "server") and self.server:
            logging.info("Cleaning up model and tokenizer resources...")

            if hasattr(self.server, "model"):
                self.server.model = None

            if hasattr(self.server, "tokenizer"):
                self.server.tokenizer = None

            if hasattr(self.server, "llm_loaded"):
                self.server.llm_loaded = None

        # Clean up local references
        if hasattr(self, "model"):
            del self.model
        if hasattr(self, "tokenizer"):
            del self.tokenizer



import json
import os
from typing import Optional
import shutil
import huggingface_hub
from importlib.metadata import distributions
from lemonade_server.pydantic_models import PullConfig
from lemonade.cache import DEFAULT_CACHE_DIR

USER_MODELS_FILE = os.path.join(DEFAULT_CACHE_DIR, "user_models.json")


class ModelManager:

    @staticmethod
    def parse_checkpoint(checkpoint: str) -> tuple[str, str | None]:
        """
        Parse a checkpoint string that may contain a variant separated by a colon.

        For GGUF models, the format is "repository:variant" (e.g., "unsloth/Qwen3-0.6B-GGUF:Q4_0").
        For other models, there is no variant.

        Args:
            checkpoint: The checkpoint string, potentially with variant

        Returns:
            tuple: (base_checkpoint, variant) where variant is None if no colon is present
        """
        if ":" in checkpoint:
            base_checkpoint, variant = checkpoint.split(":", 1)
            return base_checkpoint, variant
        return checkpoint, None

    @property
    def supported_models(self) -> dict:
        """
        Returns a dictionary of supported models.
        Note: Models must be downloaded before they are locally available.
        """
        # Load the models dictionary from the built-in JSON file
        server_models_file = os.path.join(
            os.path.dirname(__file__), "server_models.json"
        )
        with open(server_models_file, "r", encoding="utf-8") as file:
            models: dict = json.load(file)

        # Load the user's JSON file, if it exists, and merge into the models dict
        if os.path.exists(USER_MODELS_FILE):
            with open(USER_MODELS_FILE, "r", encoding="utf-8") as file:
                user_models: dict = json.load(file)
            # Prepend the user namespace to the model names
            user_models = {
                f"user.{model_name}": model_info
                for model_name, model_info in user_models.items()
            }

            # Backwards compatibility for user models that were created before version 8.0.4
            # "reasoning" was a boolean, but as of 8.0.4 it became a label
            for _, model_info in user_models.items():
                if "reasoning" in model_info:
                    model_info["labels"] = (
                        ["reasoning"]
                        if not model_info["labels"]
                        else model_info["labels"] + ["reasoning"]
                    )
                    del model_info["reasoning"]

            models.update(user_models)

        # Add the model name as a key in each entry, to make it easier
        # to access later

        for key, value in models.items():
            value["model_name"] = key

        return models

    @property
    def downloaded_hf_checkpoints(self) -> list[str]:
        """
        Returns a list of Hugging Face checkpoints that have been downloaded.
        """
        downloaded_hf_checkpoints = []
        try:
            hf_cache_info = huggingface_hub.scan_cache_dir()
            downloaded_hf_checkpoints = [entry.repo_id for entry in hf_cache_info.repos]
        except huggingface_hub.CacheNotFound:
            pass
        except Exception as e:  # pylint: disable=broad-exception-caught
            print(f"Error scanning Hugging Face cache: {e}")
        return downloaded_hf_checkpoints

    @property
    def downloaded_models(self) -> dict:
        """
        Returns a dictionary of locally available models.
        """
        downloaded_models = {}
        downloaded_checkpoints = self.downloaded_hf_checkpoints
        for model in self.supported_models:
            base_checkpoint = self.parse_checkpoint(
                self.supported_models[model]["checkpoint"]
            )[0]
            if base_checkpoint in downloaded_checkpoints:
                downloaded_models[model] = self.supported_models[model]
        return downloaded_models

    @property
    def downloaded_models_enabled(self) -> dict:
        """
        Returns a dictionary of locally available models that are enabled by
        the current installation.
        """
        return self.filter_models_by_backend(self.downloaded_models)

    def identify_gguf_models(
        self, checkpoint: str, variant: str, mmproj: str
    ) -> tuple[dict, list[str]]:
        """
        Identifies the GGUF model files in the repository that match the variant.
        """

        hint = """
        The CHECKPOINT:VARIANT scheme is used to specify model files in Hugging Face repositories.

        The VARIANT format can be one of several types:
        1. Full filename: exact file to download
        2. None/empty: gets the first .gguf file in the repository (excludes mmproj files)
        3. Quantization variant: find a single file ending with the variant name (case insensitive)
        4. Folder name: downloads all .gguf files in the folder that matches the variant name (case insensitive)
        
        Examples:
        - "unsloth/Qwen3-8B-GGUF:qwen3.gguf" -> downloads "qwen3.gguf"
        - "unsloth/Qwen3-30B-A3B-GGUF" -> downloads "Qwen3-30B-A3B-GGUF.gguf"
        - "unsloth/Qwen3-8B-GGUF:Q4_1" -> downloads "Qwen3-8B-GGUF-Q4_1.gguf"
        - "unsloth/Qwen3-30B-A3B-GGUF:Q4_0" -> downloads all files in "Q4_0/" folder
        """

        repo_files = huggingface_hub.list_repo_files(checkpoint)
        sharded_files = []

        # (case 1) If variant ends in .gguf, use it directly
        if variant and variant.endswith(".gguf"):
            variant_name = variant
            if variant_name not in repo_files:
                raise ValueError(
                    f"File {variant} not found in Hugging Face repository {checkpoint}. {hint}"
                )
        # (case 2) If no variant is provided, get the first .gguf file in the repository
        elif variant is None:
            all_variants = [
                f for f in repo_files if f.endswith(".gguf") and "mmproj" not in f
            ]
            if len(all_variants) == 0:
                raise ValueError(
                    f"No .gguf files found in Hugging Face repository {checkpoint}. {hint}"
                )
            variant_name = all_variants[0]
        else:
            # (case 3) Find a single file ending with the variant name (case insensitive)
            end_with_variant = [
                f
                for f in repo_files
                if f.lower().endswith(f"{variant}.gguf".lower())
                and "mmproj" not in f.lower()
            ]
            if len(end_with_variant) == 1:
                variant_name = end_with_variant[0]
            elif len(end_with_variant) > 1:
                raise ValueError(
                    f"Multiple .gguf files found for variant {variant}, but only one is allowed. {hint}"
                )
            # (case 4) Check whether the variant corresponds to a folder with sharded files (case insensitive)
            else:
                sharded_files = [
                    f
                    for f in repo_files
                    if f.endswith(".gguf")
                    and f.lower().startswith(f"{variant}/".lower())
                ]

                if not sharded_files:
                    raise ValueError(
                        f"No .gguf files found for variant {variant}. {hint}"
                    )

                # Sort to ensure consistent ordering
                sharded_files.sort()

                # Use first file as primary (this is how llamacpp handles it)
                variant_name = sharded_files[0]

        core_files = {"variant": variant_name}

        # If there is a mmproj file, add it to the patterns
        if mmproj:
            if mmproj not in repo_files:
                raise ValueError(
                    f"The provided mmproj file {mmproj} was not found in {checkpoint}."
                )
            core_files["mmproj"] = mmproj

        return core_files, sharded_files

    def download_gguf(self, model_config: PullConfig) -> dict:
        """
        Downloads the GGUF file for the given model configuration.

        For sharded models, if the variant points to a folder (e.g. Q4_0), all files in that folder
        will be downloaded but only the first file will be returned for loading.
        """

        # This code handles all cases by constructing the appropriate filename or pattern
        checkpoint, variant = self.parse_checkpoint(model_config.checkpoint)

        # Identify the GGUF model files in the repository that match the variant
        core_files, sharded_files = self.identify_gguf_models(
            checkpoint, variant, model_config.mmproj
        )

        # Download the files
        snapshot_folder = huggingface_hub.snapshot_download(
            repo_id=checkpoint,
            allow_patterns=list(core_files.values()) + sharded_files,
        )

        # Ensure we downloaded all expected files
        for file in list(core_files.values()) + sharded_files:
            expected_path = os.path.join(snapshot_folder, file)
            if not os.path.exists(expected_path):
                raise ValueError(
                    f"Hugging Face snapshot download for {model_config.checkpoint} "
                    f"expected file {file} not found at {expected_path}"
                )

        # Return a dict of the full path of the core GGUF files
        return {
            file_name: os.path.join(snapshot_folder, file_path)
            for file_name, file_path in core_files.items()
        }

    def download_models(
        self,
        models: list[str],
        checkpoint: Optional[str] = None,
        recipe: Optional[str] = None,
        reasoning: bool = False,
        mmproj: str = "",
    ):
        """
        Downloads the specified models from Hugging Face.
        """
        for model in models:
            if model not in self.supported_models:
                # Register the model as a user model if the model name
                # is not already registered

                # Ensure the model name includes the `user` namespace
                model_parsed = model.split(".", 1)
                if len(model_parsed) != 2 or model_parsed[0] != "user":
                    raise ValueError(
                        f"When registering a new model, the model name must "
                        "include the `user` namespace, for example "
                        f"`user.Phi-4-Mini-GGUF`. Received: {model}"
                    )

                model_name = model_parsed[1]

                # Check that required arguments are provided
                if not recipe or not checkpoint:
                    raise ValueError(
                        f"Model {model} is not registered with Lemonade Server. "
                        "To register and install it, provide the `checkpoint` "
                        "and `recipe` arguments, as well as the optional "
                        "`reasoning` and `mmproj` arguments as appropriate. "
                    )

                # JSON content that will be used for registration if the download succeeds
                new_user_model = {
                    "checkpoint": checkpoint,
                    "recipe": recipe,
                    "suggested": True,
                    "labels": ["custom"] + (["reasoning"] if reasoning else []),
                }

                if mmproj:
                    new_user_model["mmproj"] = mmproj

                # Make sure that a variant is provided for GGUF models before registering the model
                if "gguf" in checkpoint.lower() and ":" not in checkpoint.lower():
                    raise ValueError(
                        "You are required to provide a 'variant' in the checkpoint field when "
                        "registering a GGUF model. The variant is provided "
                        "as CHECKPOINT:VARIANT. For example: "
                        "Qwen/Qwen2.5-Coder-3B-Instruct-GGUF:Q4_0 or "
                        "Qwen/Qwen2.5-Coder-3B-Instruct-GGUF:"
                        "qwen2.5-coder-3b-instruct-q4_0.gguf"
                    )

                # Create a PullConfig we will use to download the model
                new_registration_model_config = PullConfig(
                    model_name=model_name,
                    checkpoint=checkpoint,
                    recipe=recipe,
                    reasoning=reasoning,
                )
            else:
                new_registration_model_config = None

            # Download the model
            if new_registration_model_config:
                checkpoint_to_download = checkpoint
                gguf_model_config = new_registration_model_config
            else:
                checkpoint_to_download = self.supported_models[model]["checkpoint"]
                gguf_model_config = PullConfig(**self.supported_models[model])
            print(f"Downloading {model} ({checkpoint_to_download})")

            if "gguf" in checkpoint_to_download.lower():
                self.download_gguf(gguf_model_config)
            else:
                huggingface_hub.snapshot_download(repo_id=checkpoint_to_download)

            # Register the model in user_models.json, creating that file if needed
            # We do this registration after the download so that we don't register
            # any incorrectly configured models where the download would fail
            if new_registration_model_config:
                if os.path.exists(USER_MODELS_FILE):
                    with open(USER_MODELS_FILE, "r", encoding="utf-8") as file:
                        user_models: dict = json.load(file)
                else:
                    user_models = {}

                user_models[model_name] = new_user_model

                # Ensure the cache directory exists before writing the file
                os.makedirs(os.path.dirname(USER_MODELS_FILE), exist_ok=True)

                with open(USER_MODELS_FILE, mode="w", encoding="utf-8") as file:
                    json.dump(user_models, fp=file)

    def filter_models_by_backend(self, models: dict) -> dict:
        """
        Returns a filtered dict of models that are enabled by the
        current environment.
        """
        installed_packages = {dist.metadata["Name"].lower() for dist in distributions()}

        hybrid_installed = (
            "onnxruntime-vitisai" in installed_packages
            and "onnxruntime-genai-directml-ryzenai" in installed_packages
        )
        filtered = {}
        for model, value in models.items():
            if value.get("recipe") == "oga-hybrid":
                if hybrid_installed:
                    filtered[model] = value
            else:
                filtered[model] = value
        return filtered

    def delete_model(self, model_name: str):
        """
        Deletes the specified model from local storage.
        """
        if model_name not in self.supported_models:
            raise ValueError(
                f"Model {model_name} is not supported. Please choose from the following: "
                f"{list(self.supported_models.keys())}"
            )

        checkpoint = self.supported_models[model_name]["checkpoint"]
        print(f"Deleting {model_name} ({checkpoint})")

        # Handle GGUF models that have the format "checkpoint:variant"
        base_checkpoint = self.parse_checkpoint(checkpoint)[0]

        try:
            # Get the local path using snapshot_download with local_files_only=True
            snapshot_path = huggingface_hub.snapshot_download(
                repo_id=base_checkpoint, local_files_only=True
            )

            # Navigate up to the model directory (parent of snapshots directory)
            model_path = os.path.dirname(os.path.dirname(snapshot_path))

            # Delete the entire model directory (including all snapshots)
            if os.path.exists(model_path):
                shutil.rmtree(model_path)
                print(f"Successfully deleted model {model_name} from {model_path}")
            else:
                raise ValueError(
                    f"Model {model_name} not found locally at {model_path}"
                )

        except Exception as e:
            if (
                "not found in cache" in str(e).lower()
                or "no such file" in str(e).lower()
            ):
                raise ValueError(f"Model {model_name} is not installed locally")
            else:
                raise ValueError(f"Failed to delete model {model_name}: {str(e)}")


# This file was originally licensed under Apache 2.0. It has been modified.
# Modifications Copyright (c) 2025 AMD



import os
import shutil
import glob
import pathlib
from typing import Dict, List, Optional
import yaml
import lemonade.common.printing as printing
import lemonade.common.build as build
import lemonade.common.exceptions as exp

CACHE_MARKER = ".lemonadecache"
BUILD_MARKER = ".lemonadebuild"


def rmdir(folder, excludes: Optional[List[str]] = None):
    """
    Remove the contents of a directory from the filesystem.
    If `<name>` is in `excludes`, the directory itself and the file named <name>
    are kept. Otherwise, the entire directory is removed.
    """

    # Use an empty list by default
    if excludes:
        excludes_to_use = excludes
    else:
        excludes_to_use = []

    if os.path.isdir(folder):
        for filename in os.listdir(folder):
            file_path = os.path.join(folder, filename)
            if file_path not in excludes_to_use:
                if os.path.isfile(file_path) or os.path.islink(file_path):
                    os.unlink(file_path)
                elif os.path.isdir(file_path):
                    shutil.rmtree(file_path)

        if excludes is None:
            shutil.rmtree(folder)

        return True

    else:
        return False


def get_all(path, exclude_path=False, file_type=build.state_file_name, recursive=True):
    if recursive:
        files = [
            os.path.join(dp, f)
            for dp, dn, filenames in os.walk(path)
            for f in filenames
            if file_type in f
        ]
    else:
        files = []
        dp, _, filenames = os.walk(path)
        for f in filenames:
            if file_type in f:
                files.append(os.path.join(dp, f))

    if exclude_path:
        files = [os.path.basename(f) for f in files]

    return files


def clean_file_name(script_path: str) -> str:
    """
    Trim the ".py" / ".onnx" if present.

    If its a state.yaml file, trim the "_state.yaml"
    """

    if script_path.endswith("_" + build.state_file_name):
        return pathlib.Path(script_path).stem.replace(
            "_" + os.path.splitext(build.state_file_name)[0], ""
        )
    else:
        return pathlib.Path(script_path).stem


class CacheError(exp.Error):
    """
    Raise this exception when the cache is being accessed incorrectly
    """


def _load_yaml(file) -> Dict:
    if os.path.isfile(file):
        with open(file, "r", encoding="utf8") as stream:
            return yaml.load(stream, Loader=yaml.FullLoader)
    else:
        return {}


def save_yaml(dict: Dict, file):
    with open(file, "w", encoding="utf8") as outfile:
        yaml.dump(dict, outfile)


def print_yaml_file(file_path, description):
    if os.path.exists(file_path):
        with open(file_path, "r", encoding="utf-8") as file:
            printing.log_info(f"The {description} for {file_path} are:")
            print(file.read())
    else:
        raise CacheError(
            f"No {description} found at {file_path}. "
            "Try running `lemonade cache --list` to see the builds in your build cache."
        )


def make_cache_dir(cache_dir: str):
    """
    Create the build and cache directories, and put hidden files in them
    to mark them as such.
    """

    os.makedirs(cache_dir, exist_ok=True)

    # File that indicates that the directory is a cache directory
    cache_file_path = os.path.join(cache_dir, CACHE_MARKER)
    open(cache_file_path, mode="w", encoding="utf").close()


def make_build_dir(cache_dir: str, build_name: str):
    """
    Create the build and cache directories, and put hidden files in them
    to mark them as such.
    """
    make_cache_dir(cache_dir)

    build_dir = build.output_dir(cache_dir, build_name)
    os.makedirs(build_dir, exist_ok=True)

    # File that indicates that the directory is a build directory
    build_file_path = os.path.join(build_dir, BUILD_MARKER)
    open(build_file_path, mode="w", encoding="utf").close()


def check_cache_dir(cache_dir: str):
    cache_file_path = os.path.join(cache_dir, CACHE_MARKER)
    if not os.path.isfile(cache_file_path):
        raise CacheError(
            f"{cache_dir} is not a cache directory generated by Lemonade. "
            "You can only clean, delete and generate reports for directories that "
            "have been generated by Lemonade. Set a different --cache-dir before "
            "trying again."
        )


def is_build_dir(cache_dir: str, build_name: str):
    build_dir = build.output_dir(cache_dir, build_name)
    build_file_path = os.path.join(build_dir, BUILD_MARKER)
    return os.path.isfile(build_file_path)


def clean_output_dir(cache_dir: str, build_name: str) -> None:
    """
    Delete all elements of the output directory that are not human readable
    """
    output_dir = build.output_dir(cache_dir, build_name)
    if os.path.isdir(output_dir) and is_build_dir(cache_dir, build_name):
        output_dir = os.path.expanduser(output_dir)
    else:
        raise CacheError(f"No build found at {output_dir}")

    # Remove files that do not have an allowed extension
    allowed_extensions = (".txt", ".out", ".yaml", ".json", ".png")
    all_paths = glob.glob(f"{output_dir}/**/*", recursive=True)
    for path in all_paths:
        if os.path.isfile(path) and not path.endswith(allowed_extensions):
            os.remove(path)

    # Remove all empty folders
    for path in all_paths:
        if os.path.isdir(path):
            if len(os.listdir(path)) == 0:
                shutil.rmtree(path)


def get_available_builds(cache_dir):
    """
    Get all of the build directories within the build cache
    located at `cache_dir`
    """

    check_cache_dir(cache_dir)

    builds = [
        pathlib.PurePath(build_name).name
        for build_name in os.listdir(os.path.abspath(build.builds_dir(cache_dir)))
        if os.path.isdir(build.output_dir(cache_dir, build_name))
        and is_build_dir(cache_dir, build_name)
    ]
    builds.sort()

    return builds


class Keys:
    # Number of parameters in the model
    PARAMETERS = "parameters"
    # List of all build tools in the Sequence
    SELECTED_SEQUENCE_OF_TOOLS = "selected_sequence_of_tools"
    # MeasuredPerformance data for a benchmarked workload
    PERFORMANCE = "performance"
    # Runtime used for the benchmark
    RUNTIME = "runtime"
    # Type of device used for the benchmark (e.g., "x86")
    DEVICE_TYPE = "device_type"
    # Specific device used for the benchmark
    DEVICE = "device"
    # Name of the model
    MODEL_NAME = "model_name"
    # Number of iterations used in benchmarking
    ITERATIONS = "iterations"
    # System information to keep track of DUT
    SYSTEM_INFO = "system_info"
    # Indicates status of the most recent build tool run: FunctionStatus
    BUILD_STATUS = "build_status"
    # Prefix for reporting the execution duration of a tool
    # In the report this will look like tool_duration:TOOL_NAME
    TOOL_DURATION = "tool_duration"
    # Prefix for reporting the peak working memory in the build through this tool
    # In the report this will look like tool_memory:TOOL_NAME
    TOOL_MEMORY = "tool_memory"
    # Prefix for reporting the execution status of a tool
    # In the report this will look like tool_status:TOOL_NAME
    TOOL_STATUS = "tool_status"
    # Records the date and time of the evaluation after analysis but before
    # build and benchmark
    TIMESTAMP = "timestamp"
    # Records the logfile of any failed tool/benchmark
    ERROR_LOG = "error_log"
    # Name of the build in the cache
    BUILD_NAME = "build_name"
    # Sequence of tools used for this build, along with their args
    SEQUENCE_INFO = "sequence_info"
    # Version of Lemonade used for the build
    LEMONADE_VERSION = "lemonade_version"
    # Unique ID for this build
    UID = "uid"
    # Directory where the lemonade build cache is stored
    CACHE_DIR = "cache_dir"
    # Example inputs to the model
    INPUTS = "inputs"
    # Path to the file containing the memory usage plot
    MEMORY_USAGE_PLOT = "memory_usage_plot"
    # Average of all tested MMLU subject scores
    AVERAGE_MMLU_ACCURACY = "average_mmlu_accuracy"


def _clean_logfile(logfile_lines: List[str]) -> List[str]:
    """
    Remove the whitespace and empty lines from an array of logfile lines
    """
    return "\n".join([line.rstrip() for line in logfile_lines if line.rstrip()])


def stats_file(cache_dir: str, build_name: str):
    """
    Returns the expected location of the lemonade stats file
    """
    dir = build.output_dir(cache_dir, build_name)
    return os.path.join(dir, "lemonade_stats.yaml")


class Stats:
    def __init__(self, cache_dir: str, build_name: str):
        self.file = stats_file(cache_dir, build_name)

        os.makedirs(os.path.dirname(self.file), exist_ok=True)
        if not os.path.exists(self.file):
            # Start an empty stats file
            save_yaml({}, self.file)

    @property
    def stats(self):
        return _load_yaml(self.file)

    def _set_key(self, dict, keys: List["str"], value):
        """
        Recursive approach to safely setting a key within any level of hierarchy
        in a dictionary. If a parent key of the desired key does not exist, create
        it and set it with an empty dictionary before proceeding.

        The end result is: dict[keys[0]][keys[1]]...[keys[-1]] = value
        """
        if len(keys) == 1:
            dict[keys[0]] = value

        else:
            if keys[0] not in dict.keys():
                dict[keys[0]] = {}

            self._set_key(dict[keys[0]], keys[1:], value)

    def save_stat(self, key: str, value):
        """
        Save statistics to an yaml file in the build directory
        """

        stats_dict = self.stats

        self._set_key(stats_dict, [key], value)

        save_yaml(stats_dict, self.file)

    def save_sub_stat(self, parent_key: str, key: str, value):
        stats_dict = self.stats

        self._set_key(stats_dict, [parent_key, key], value)

        save_yaml(stats_dict, self.file)

    def save_eval_error_log(self, logfile_path):
        if logfile_path is None:
            # Avoid an error in the situation where we crashed before
            # initializing the tool (in which case it has no logfile path yet)
            return
        if os.path.exists(logfile_path):
            with open(logfile_path, "r", encoding="utf-8") as f:
                full_log = f.readlines()

                # Log files can be quite large, so we will just record the beginning
                # and ending lines. Users can always open the log file if they
                # want to see the full log.
                start_cutoff = 5
                end_cutoff = -30
                max_full_length = start_cutoff + abs(end_cutoff)

                if len(full_log) > max_full_length:
                    log_start = _clean_logfile(full_log[:start_cutoff])
                    log_end = _clean_logfile(full_log[end_cutoff:])
                    truncation_notice = (
                        "NOTICE: This copy of the log has been truncated to the first "
                        f"{start_cutoff} and last {abs(end_cutoff)} lines "
                        f"to save space. Please see {logfile_path} "
                        "to see the full log.\n"
                    )

                    stats_log = log_start + truncation_notice + log_end
                else:
                    stats_log = _clean_logfile(full_log)

                self.save_stat(Keys.ERROR_LOG, stats_log)


def expand_inputs(input_paths: List[str]) -> List[str]:
    """
    Convert regular expressions in input paths
    into full file/dir paths (e.g., [*.py] -> [a.py, b.py] )

    This makes up for Windows not resolving wildcards on the command line
    """
    input_paths_expanded = sum(
        [glob.glob(f) for f in input_paths if "::" not in f], []
    ) + [f for f in input_paths if "::" in f]

    if not input_paths_expanded:
        raise exp.ArgError("No files that match your inputs could be found.")

    return input_paths_expanded


# This file was originally licensed under Apache 2.0. It has been modified.
# Modifications Copyright (c) 2025 AMD
